#! /usr/bin/env python3

# NEED to arrive at a consensus for protein and transcript IDs

import re
import sys
import argparse
from mycotools.lib.biotools import gff2list, list2gff, gff2Comps, gff3Comps
from mycotools.lib.kontools import format_path, eprint, vprint
from mycotools.curAnnotation import addGenes, removeStartStop

def gff2gff3(gff_list, ome, jgi_ome):

    comps2, exon_dict, cds_dict, out_list, gene_dict = gff2Comps(), {}, {}, [], {}
    for entry in gff_list:
        if entry['type'] == 'exon':
            name = re.search( comps2['id'], entry['attributes'] )[1]
            transcript = re.search( comps2['transcript'], entry['attributes'] )[1]
            if name not in exon_dict:
                exon_dict[name] = 0
            if name not in gene_dict:
                gene_dict[name] = { 
                    'protein': '', 'transcript': '', 'product': ''
                    }
            gene_dict[name]['transcript'] = transcript
            exon_dict[name] += 1
            exon_id = 'ID=exon_' + transcript + '_' + str(exon_dict[name])
            par_id = 'mRNA_' + transcript
            entry['attributes'] = exon_id + ';Parent=' + par_id + ';Alias=' + name
        elif entry['type'] == 'CDS':
            name = re.search( comps2['id'], entry['attributes'] )[1]
            protein = re.search( comps2['prot'], entry['attributes'] )[1]
            product = re.search( comps2['product'], entry['attributes'] )
            if name not in cds_dict:
                cds_dict[name] = 0
            if name not in gene_dict:
                gene_dict[name] = { 
                    'protein': '', 'transcript': '', 'product': ''
                    }
            if product is not None:
                gene_dict[name]['product'] = product[1]
            gene_dict[name]['protein'] = protein
            cds_dict[name] += 1
            cds_id = 'CDS_$_' + str( cds_dict[name] )
            entry['attributes'] = 'ID=' + cds_id + ';Alias=' + name

    comps3 = gff3Comps()
    for entry in gff_list:
        if entry['type'] not in {'start_codon', 'stop_codon'}:
            if entry['type'] == 'exon':
                name = re.search( comps3['Alias'], entry['attributes'] )[1]
                trans = gene_dict[name]['transcript']
                prot = ome + '_' + gene_dict[name]['protein']
                entry['attributes'] = re.sub(
                    comps3['Alias'], 
                    'Parent=mRNA_' + trans + ';Alias=' + prot, 
                    entry['attributes']
                    )
            elif entry['type'] == 'CDS':
                name = re.search( comps3['Alias'], entry['attributes'] )[1]
                trans = gene_dict[name]['transcript']
                prot = ome + '_' + gene_dict[name]['protein']
                entry['attributes'] = re.sub(
                    r'ID=CDS_\$_(\d+)', 
                    'ID=CDS_' + trans + r'_\1;' + 'Parent=mRNA_' + trans,
                    entry['attributes']
                    )
                entry['attributes'] = re.sub(
                    comps3['Alias'], 'Alias=' + prot, entry['attributes']
                    )
            elif entry['type'] == 'gene':
                name = re.search( comps2['id'], entry['attributes'] )[1]
                gene = 'gene_' + gene_dict[name]['transcript']
                prot = gene_dict[name]['protein']
                jgi = 'jgi.p|' + jgi_ome + '|' + prot
                prod = gene_dict[name]['product']
                trans = gene_dict[name]['transcript']
                alias = ome + '_' + prot
                entry['attributes'] = 'ID=' + gene + ';Name=' + jgi + \
                    ';portal_id=' + jgi_ome + ';product_name=' + prod + \
                    ';proteinId=' + prot + ';transcriptId=' + trans + \
                    ';Alias=' + alias
            elif entry['type'] == 'mRNA':
                name = re.search( comps2['id'], entry['attributes'] )[1]
                gene = 'gene_' + gene_dict[name]['transcript']
                mrna = 'mRNA_' + gene_dict[name]['transcript']
                prot = gene_dict[name]['protein']
                jgi = 'jgi.p|' + jgi_ome + '|' + prot
                trans = gene_dict[name]['transcript']
                alias = ome + '_' + prot
                entry['attributes'] = 'ID=' + mrna + ';Name=' + jgi + \
                    ';Parent=' + gene + \
                    ';proteinId=' + prot + ';transcriptId=' + trans + \
                    ';Alias=' + alias
            out_list.append( entry )

    return out_list


def main(gff_list, ome, jgi_ome, safe = True, verbose = True,
         cur_seqids = False):

    gff_prep, failed, flagged = addGenes( gff_list, safe = safe )
    if failed:
        vprint( str(len(failed)) + '\tgenes failed', v = verbose , e = True, flush = True)
    if flagged:
        vprint( str(len(flagged)) + '\tgene coordinates from exons', v = verbose, e = True, flush = True)
    gff3 = gff2gff3(gff_prep, ome, jgi_ome)
    if cur_seqids:
        for entry in gff3:
            if not entry['seqid'].startswith(ome + '_'):
                entry['seqid'] = ome + '_' + entry['seqid']

    return gff3

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser( description = 'Converts jgi gff2 to gff3' )
    parser.add_argument( '-i', '--input', required = True, help = 'JGI gff2' )
    parser.add_argument( '-o', '--ome', required = True, help = 'Internal ome' )
    parser.add_argument( '-j', '--jgi', required = True, help = 'JGI ome' )
    parser.add_argument( '--fail', default = True, action = 'store_false', \
        help = 'Fail genes w/o CDS sequences that lack start or stop codons' )
    args = parser.parse_args()

    gff_list = gff2list(format_path( args.input ))
    eprint( args.ome + '\t' + args.input , flush = True)
    gff3 = main( gff_list, args.ome, args.jgi, args.fail )
    print( list2gff( gff3 ) , flush = True)

    sys.exit( 0 )



#! /usr/bin/env python3

# NEED to curate CDS parents to ensure they relate to the RNA
# NEED to curate protein ids to be a uniform title
# NEED to fix mobile_genetic_element conversion from NCBI
    # entries contain a gene, mRNA, exon
# NEED to restrict passing types [mobile_genetic_element], 
    # or make them conform
        # change alias to accomodate these edge cases:
            # Long_terminal_repeat, mobile_genetic_element
                # MRNA should not include a CDS here
# NEED to fix standalone gene alias
    # an alias is applied like "Alias" with nothing else
    # only sometimes
        # if followed by unreviewed type

import os
import re
import sys
import copy
from collections import defaultdict
from mycotools.lib.kontools import format_path, sys_start, eprint
from mycotools.lib.biotools import gff2list, list2gff, gff3Comps

class RNAError(Exception):
    pass

def addMissing(gff_list, intron, comps, ome):
    out_genes, t_list, rnas, introns = {}, [], {}, {}
    mtdb_count, pseudocount, alt_alias = 1, 1, {}
    rna_changes = {} # a dictionary for changing ambigious rna id names for
    # explicit RNA type references
    for entry in gff_list:
        addEntry = None
        id_ = re.search(comps['id'], entry['attributes'])[1]
        if 'gene' in entry['type']: # includes pseudogenes
            out_genes[id_] = {
                'gene': [entry], 'tmrna': [], 'rna': [], 
                'cds': [], 'exon': [], 'texon': [], 
                'etc': [], 'pseudo': entry['type'] == 'pseudogene'
                } # create an entry for the genes
            if 'gene_biotype=protein_coding' in entry['attributes']: # if it is
            # NCBI protein coding gene (or conforming format)
                addEntry = copy.deepcopy(entry)
                addEntry['type'] = 'mRNA'
                addEntry['attributes'] = addEntry['attributes'].replace('ID=gene-', 'ID=mrna-')
                addEntry['attributes'] += ';Parent=' + id_
                addEntry['attributes'] = addEntry['attributes'].replace('gbkey=Gene', 'gbkey=mRNA')
                addEntry['attributes'] = addEntry['attributes'].replace('gene_biotype=protein_coding;','')
                out_genes[id_]['tmrna'].append(addEntry) # add a temporary RNA
                # in case there isn't one
        elif entry['type'] == 'CDS':
            try: # attempt to acquire the parent
                par = re.search(comps['par'], entry['attributes'])[1]
            except TypeError:
                continue # if there isn't a parent then skip it
            if par in rna_changes:
                entry['attributes'] = re.sub(comps['par'], 'Parent=' \
                                             + rna_changes[par],
                                             entry['attributes'])
                par = rna_changes[par]
            if par in rnas: # if the parent is an RNA, acquire the gene
                rna_par = par
                par = rnas[par] # change par to gene
            else:
                rna_par = None
            search = re.search(comps['prot'], entry['attributes'])
            if search: # search for protein
                prot = search.groups()[0] # try the first search
                if not prot:
                    prot = search.groups()[1] # try the second
                if re.search(comps['Alias'], entry['attributes'] ) is None:
                # is there a mycotools alias or associated alias?
                    if not entry['attributes'].endswith( ';' ):
                        entry['attributes'] += ';'
                    alias = ome + '_' + prot # create an alias with the found
                    # protein
                    entry['attributes'] += 'Alias=' + ome + '_' + prot
                    alt_alias[rna_par] = ome + '_' + prot
            else: # doesnt have a protein id
                if rna_par: # if there is an RNA
                    if rna_par not in alt_alias: # add it in
                        alias = ome + '_mtdb' + str(mtdb_count)
                        mtdb_count += 1
                        alt_alias[rna_par] = alias
                    else: # acquire the alias that is used
                        alias = alt_alias[rna_par]
                else: # no RNA parent
                    if par not in alt_alias:
                        if out_genes[par]['pseudo']: # if it is a pseudogene
                            alias = ome + '_pseudogene' + str(pseudocount)
                            pseudocount += 1
                        else:
                            alias = ome + '_mtdb' + str(mtdb_count)
                            mtdb_count += 1
                        alt_alias[par] = alias
                    else:
                        alias = alt_alias[par]
                if not entry['attributes'].endswith(';'):
                    entry['attributes'] += ';'
                entry['attributes'] += 'Alias=' + alias

            if not out_genes[par]['pseudo']: # if it isnt a defined pesudogene
                entry['attributes'] = entry['attributes'].replace('Parent=gene-', 'Parent=mrna-')
                out_genes[par]['cds'].append(entry)
                if not intron:
                    addEntry = copy.deepcopy(entry)
                    addEntry['type'] = 'exon'
                    addEntry['attributes'] = addEntry['attributes'].replace('ID=cds-', 'ID=exon-')
                    addEntry['attributes'] = addEntry['attributes'].replace('gbkey=CDS', 'gbkey=mRNA')
                    out_genes[par]['texon'].append(addEntry)
            else: # else just add the unmodified CDS
                out_genes[par]['cds'].append(entry)
        elif 'RNA' in entry['type'] or entry['type'] == 'transcript':
            entry['type'] = entry['type'].replace('transcript','RNA')
            if id_.startswith('rna'): # make RNA ID explicit
                new_id = entry['type'].lower() + id_[3:]
                entry['attributes'] = re.sub(comps['id'], 'ID=' + new_id,
                                             entry['attributes'])
                rna_changes[id_] = new_id
                id_ = new_id
            try:
                par = re.search(comps['par'], entry['attributes'])[1]
            except TypeError: # no gene in the ids, initialize new entry
                addEntry = copy.deepcopy(entry)
                addEntry['type'] = 'gene'
                geneID = id_.replace('rna-', 'gene-')
                if geneID == id_:
                    raise RNAError('RNA ID annotated as gene')
                addEntry['attributes'] = addEntry['attributes'].replace('ID=rna-', 'ID=gene-')
                addEntry['attributes'] = addEntry['attributes'].replace(entry['type'], 'gene')
                addEntry['attributes'] = re.sub(r'gbkey=[^;]+', 'gbkey=gene',
                                                addEntry['attributes'])
                addEntry['attributes'] = re.sub(comps['par'] + ';', '',
                                                addEntry['attributes'])
                                                # remove parent
                entry['attributes'] = re.sub(comps['par'] + ';', '',
                                             entry['attributes'])
                out_genes[geneID] = {
                    'gene': [addEntry], 'tmrna': [], 'rna': [], 'cds': [], 
                    'exon': [], 'texon': [], 'etc': [] 
                    }
#                out_genes[geneID]['gene'].append(addEntry)
                entry['attributes'] += ';Parent=' + geneID
                par = re.search(comps['par'], entry['attributes'])[1]

            rnas[id_] = par
            out_genes[par]['rna'].append(entry)
        elif entry['type'] == 'intron':
            try:
                par = re.search(comps['par'], entry['attributes'])[1]
            except TypeError:
                continue
            if par not in introns:
                introns[par] = []
            introns[par].append(entry)
        else:
            try:
                par = re.search(comps['par'], entry['attributes'])[1]
            except TypeError: # skip no parents, should be a flag
                continue
            if entry['type'] == 'exon':
                if par in rna_changes:
                    entry['attributes'] = re.sub(comps['par'], 'Parent=' \
                                                 + rna_changes[par],
                                                 entry['attributes'])
                    par = rna_changes[par]
                try:
                    out_genes[rnas[par]]['exon'].append(entry)
                except KeyError:
                    entry['attributes'] = entry['attributes'].replace('Parent=gene-', 'Parent=mrna-')
                    # assumes it is derived from an mRNA
                    out_genes[par]['exon'].append(entry)
            else:
                try:
                    out_genes[par]['etc'].append(entry)
                except KeyError: # derived from an RNA
                    out_genes[rnas[par]]['etc'].append(entry)

    if introns:
        exons = {}
        for gene, intronList in introns.items():
            if out_genes[gene]['exon']:
                continue
            geneCoords = sorted([int(out_genes[gene]['gene'][0]['start']), int(out_genes[gene]['gene'][0]['end'])])
            intronList = sorted(intronList, key = lambda x: int(x['start']))
            intronCoords = [sorted([int(x['start']), int(x['end'])]) for x in intronList]
            if len(intronCoords) == 1:
                exonCoords0 = [geneCoords[0], intronCoords[0][1] - 1]
                exonCoords1 = [intronCoords[0][1] + 1, geneCoords[1]]
                addEntry = copy.deepcopy(intronList[0])
                addEntry['attributes'] = addEntry['attributes'].replace('gbkey=intron', 'gbkey=mRNA')
                addEntry['attributes'] = addEntry['attributes'].replace('ID=intron-', 'ID=exon-')
                addEntry['type'] = 'exon'
                addEntry['start'], addEntry['end'] = str(exonCoords0[0]), str(exonCoords0[1])
                out_genes[gene]['exon'].append(addEntry)
                addEntry1 = copy.deepcopy(addEntry)
                addEntry1['start'], addEntry1['end'] = str(exonCoords1[0]), str(exonCoords1[1])
                out_genes[gene]['exon'].append(addEntry1)
            else:
                for i, entryCoords in enumerate(intronCoords):
                    if i == 0:
                        exonCoords = [geneCoords[0], entryCoords[0] - 1]
                    elif i == len(intronCoords) - 1:
                        exonCoords = [entryCoords[1] + 1, geneCoords[1]]
                    else:
                        exonCoords = [intronCoords[i-1][1]+1, intronCoords[i][0] - 1]
                    addEntry = copy.deepcopy(intronList[i])
                    addEntry['attributes'] = addEntry['attributes'].replace('gbkey=intron', 'gbkey=mRNA')
                    addEntry['attributes'] = addEntry['attributes'].replace('ID=intron-', 'ID=exon-')
                    addEntry['type'] = 'exon'
                    addEntry['start'], addEntry['end'] = str(exonCoords[0]), str(exonCoords[1])
                    out_genes[gene]['exon'].append(addEntry)
                
    out_list = []
    for geneID, geneInfo in out_genes.items():
        multiRNA = False
        if geneInfo['rna']:
#            if geneInfo['rna'][0]['type'] != 'mRNA' and not geneInfo['cds']:
 #               del geneInfo['tmrna']
            if len(geneInfo['rna']) > 1:
                multiRNA = True
            if any(x['type'] == 'mRNA' for x in geneInfo['rna']):
                del geneInfo['tmrna']
        elif geneInfo['tmrna'] and not geneInfo['cds']:
            del geneInfo['tmrna']
        if geneInfo['texon'] and geneInfo['exon']:
            if multiRNA:
                exonCoords0 = set(tuple(sorted([int(x['start']), int(x['end'])])) for x in geneInfo['exon'])
                exonCoords1 = set(tuple(sorted([int(x['start']), int(x['end'])])) for x in geneInfo['texon'])
                if all(x in exonCoords0 for x in exonCoords1):
                    del geneInfo['texon']
            else:
                del geneInfo['texon']
        for seqType, seqEntry in geneInfo.items():
            if seqType != 'pseudo':
                out_list.extend(seqEntry)

    for entry in out_list: # this will relate modified RNA IDs from
    # nonsequential gffs post hoc. not sure if there are more downstream
    # implications to changing RNAs after the above curation
        try:
            par = re.search(comps['par'], entry['attributes'])[1]
        except TypeError:
            continue
        if par in rna_changes:
            new_par = rna_changes[par]
            entry['attributes'] = re.sub(comps['par'], 'Parent=' + new_par,
                                         entry['attributes'])
    return out_list, pseudocount

def acquireFormat( gff_list ):

    prot_comp = re.compile( r'ID=(.*?)[;|$]' )
    gene = False
    for line in gff_list:
        if line['type'] == 'gene':
            gene = True
            break

    if gene:
        if prot_comp.search( line['attributes'] ):
            if 'Name=jgi.p' in line['attributes']:
                return 'jgi_gff3'
            else:
                return 'misc_gff3'
    return None

def compileGenes(cur_list, ome, pseudocount = 0, comps = gff3Comps(),
                 cur_seqids = False):

    genes, pseudogenes, rnas = [], [], {}
    mrnas = defaultdict(list) 
    trnas = defaultdict(list)
    rrnas = defaultdict(list)
    ptrnas = defaultdict(list)
    transcripts = defaultdict(list)
    for i, v in enumerate(cur_list):
        id_ = re.search(comps['id'], v['attributes'])[1]
        if 'gene' in v['type']:
            genes.append((i, id_,))
            if v['type'] == 'pseudogene':
                pseudogenes.append(id_)
        elif 'RNA' in v['type']:
            par = re.search(comps['par'], v['attributes'])[1]
            rnas[id_] = par # rnaID2geneID
            if v['type'] == 'mRNA':
                mrnas[par].append(id_)
            elif v['type'] == 'tRNA':
                trnas[par].append(id_)
            elif v['type'] == 'rRNA':
                rrnas[par].append(id_)
            elif v['type'] == 'pseudogenic_tRNA':
                ptrnas[par].append(id_)
            elif v['type'] == 'RNA':
                transcripts[par].append(id_)

    id_dict, pseudogenes = {}, set(pseudogenes)
    if cur_seqids:
        for i, entry in enumerate(cur_list):
            if not entry['seqid'].startswith(ome + '_'):
                entry['seqid'] = ome + '_' + entry['seqid']
            if 'gene' in entry['type']:
                id_ = re.search(comps['id'], entry['attributes'])[1]
                id_dict[id_] = {}
            elif entry['type'] == 'CDS':
                try:
                    par = re.search(comps['par'], entry['attributes'])[1]
                except TypeError:
                    continue
                try:
                    id_dict[rnas[par]][par] = re.search(comps['Alias'], entry['attributes'])[1]
                except KeyError: # pseudogene or CDS parent is related to gene, not
                # RNA
                    id_dict[par][par] = re.search(comps['Alias'],
                                                  entry['attributes'])[1]
    else:
        for i, entry in enumerate(cur_list):
            if 'gene' in entry['type']:
                id_ = re.search(comps['id'], entry['attributes'])[1]
                id_dict[id_] = {}
            elif entry['type'] == 'CDS':
                try:
                    par = re.search(comps['par'], entry['attributes'])[1]
                except TypeError:
                    continue
                try:
                    id_dict[rnas[par]][par] = re.search(comps['Alias'], entry['attributes'])[1]
                except KeyError: # pseudogene or CDS parent is related to gene, not
                # RNA
                    id_dict[par][par] = re.search(comps['Alias'],
                                                  entry['attributes'])[1]

    trna_count, rrna_count, ptrna_count, rna_count, etc_count = 1, 1, 1, 1, 1
    estranged = []
    for i, entry in enumerate(cur_list):
        alias_tag, rna_check = ';Alias=', False
        if 'gene' in entry['type']:
            id_ = re.search(comps['id'], entry['attributes'])[1]
            if id_ in mrnas:
                rna_check = True
                if entry['type'] != 'pseudogene':
                    for rna,alias in id_dict[id_].items():
                        alias_tag += alias + '|'
                else:
                    for rna in mrnas[id_]:
                        alias = ome + '_pseudogene' + str(pseudocount)
                        id_dict[id_][rna] = alias
                        alias_tag += alias + '|'
                        pseudocount += 1
            if id_ in trnas:
                rna_check = True
                for rna in trnas[id_]: 
                    alias = ome + '_tRNA' + str(trna_count)
                    id_dict[id_][rna] = alias 
                    # only works if the gff list is sequential
                    alias_tag += alias + '|'
                    trna_count += 1
            if id_ in rrnas:
                rna_check = True
                for rna in rrnas[id_]:
                    alias = ome + '_rRNA' + str(rrna_count)
                    id_dict[id_][rna] = alias
                    # only works if gff list is sequential
                    alias_tag += alias + '|'
                    rrna_count += 1
            if id_ in ptrnas:
                rna_check = True
                for rna in ptrnas[id_]:
                    alias = ome + '_pseudo-tRNA' + str(ptrna_count)
                    id_dict[id_][rna] = alias
                    alias_tag += alias + '|'
                    ptrna_count += 1
            if id_ in transcripts:
                rna_check = True
                for rna in transcripts[id_]:
                    alias = ome + '_RNA' + str(rna_count)
                    id_dict[id_][rna] = alias
                    alias_tag += alias + '|'
                    rna_count += 1
            if not rna_check and id_ not in pseudogenes: # no association with RNA
                alias = ome + '_etc' + str(etc_count)
                alias_tag += alias + '|'
                etc_count += 1
            alias_tag = alias_tag[:-1]
            entry['attributes'] += alias_tag
        else:
            if 'Alias=' not in entry['attributes']:
                if 'RNA' in entry['type']:
                    id_ = re.search(comps['id'], entry['attributes'])[1]
                    try:
                        alias_tag += id_dict[rnas[id_]][id_]
                    except KeyError:
                        if id_ in rnas:
                            estranged.append((i, id_,))
                            continue
#                            raise KeyError
                        else:
                            raise RNAError('unknown RNA type: ' + entry['type'])
                else:
                    try:
                        par = re.search(comps['par'], entry['attributes'])[1]
                    except TypeError: # no parent
                        continue
                    try:
                        alias_tag += id_dict[rnas[par]][par]
                    except KeyError: # no reference gene/reference transcript
#                       eprint(re.search(comps['id'], entry['attributes'])[1], entry['type'])
                        estranged.append((i, id_,))
                        continue
                entry['attributes'] += alias_tag

    for i, id_ in estranged:
        if not 'Alias=' in entry['attributes']:
            gene = rnas[id_]
            try: # acquire from the RNA retroactively
                cur_list[i]['attributes'] += ';Alias=' + id_dict[gene][id_]
            except KeyError: # acquire from the gene
                cur_list[i]['attributes'] += ';Alias=' + id_dict[gene][gene]
                id_dict[gene][id_] = id_dict[gene][gene]
                # this assumes that for the case where CDSs' parents are genes,
                # and mRNAs thus aren't related to the CDS alias directly, then
                # the RNA can be related to the alias through just the gene;
                # however, this will overlook alternately spliced genes in this
                # rare instance. Ideally, CDS parents should be curated to the
                # RNA somehow


    return cur_list


def curGff3(gff_list, ome, cur_seqids = False):

    cur_list, intron = [], False
    for line in gff_list:
        if line['type'] == 'intron':
            intron = True

    cur_list, pseudocount = addMissing(gff_list, intron, gff3Comps(), ome)
    final_list = compileGenes(cur_list, ome, pseudocount,
                              cur_seqids = cur_seqids)

    return final_list


def main(gff_path, ome, cur_seqids = False):

    if isinstance(gff_path, str):
        gff = gff2list( format_path(gff_path) )
    elif isinstance(gff_path, list):
        gff = gff_path
    typ = acquireFormat( gff )

    if not typ:
        eprint('\tERROR: type unknown ' + gff_path, flush = True)
        return None

    new_gff = curGff3(gff, ome, cur_seqids)

    return new_gff


if __name__ == '__main__':
    usage = 'Imports gene coordinates file gff3, ome and curates headers'
    sys_start( sys.argv, usage, 3, files = [sys.argv[1]] )
    cur_gff = main( format_path(sys.argv[1]), sys.argv[2] )
    print( list2gff( cur_gff ) , flush = True)
    sys.exit( 0 )



#! /usr/bin/env python3

import os
import re
import sys
from mycotools.lib.kontools import sys_start, format_path, fmt_float
from mycotools.lib.biotools import fa2dict, calc_weight


if __name__ == '__main__':

    usage = 'USAGE: Inputs amino acid fasta outputs linear protein weights'
    args = sys_start(sys.argv, usage, 1)

    fa = fa2dict(format_path(sys.argv[1]))
    print('#protein\tkDa', flush = True)
    for head in fa:
        print(head + '\t' + fmt_float(calc_weight(fa[head]['sequence'])), flush = True)



#! /usr/bin/env python3

import os
import re
import sys
import glob
import gzip
import json
import shutil
import tarfile
import subprocess
from datetime import datetime

def stdin2str():
    data = ''
    for line in sys.stdin:
        data += line.rstrip() + '\n'
    data = data.rstrip()
    return data

def hex2rgb(hexCode):
    return tuple(int(hexCode.lstrip('#')[i:i+2], 16) for i in (0, 2, 4))


def getColors(size, ignore = [], rgb = False):
    if size < 16:
        colors = [
            "#000000","#004949","#009292","#ff6db6","#ffb6db",
            "#490092","#006ddb","#b66dff","#6db6ff","#b6dbff",
            "#920000","#924900","#db6d00","#24ff24","#ffff6d"
            ]
    elif size < 27:
        colors = [
            "#F0A3FF", "#0075DC", "#993F00", "#4C005C", "#191919",
            "#005C31", "#2BCE48", "#FFCC99", "#808080", "#94FFB5",
            "#8F7C00", "#9DCC00", "#C20088", "#003380", "#FFA405",
            "#FFA8BB", "#426600", "#FF0010", "#5EF1F2", "#00998F",
            "#E0FF66", "#740AFF", "#990000", "#FFFF80", "#FFFF00",
            "#FF5005"
            ]
    else:
        colors = [
            '#000000', '#010067', '#d5ff00', '#ff0056', '#9e008e', 
            '#0e4ca1', '#ffe502', '#005f39', '#00ff00', '#95003a', 
            '#ff937e', '#a42400', '#001544', '#91d0cb', '#620e00', 
            '#6b6882', '#0000ff', '#007db5', '#6a826c', '#00ae7e', 
            '#c28c9f', '#be9970', '#008f9c', '#5fad4e', '#ff0000', 
            '#ff00f6', '#ff029d', '#683d3b', '#ff74a3', '#968ae8', 
            '#98ff52', '#a75740', '#01fffe', '#ffeee8', '#fe8900', 
            '#bdc6ff', '#01d0ff', '#bb8800', '#7544b1', '#a5ffd2', 
            '#ffa6fe', '#774d00', '#7a4782', '#263400', '#004754', 
            '#43002c', '#b500ff', '#ffb167', '#ffdb66', '#90fb92', 
            '#7e2dd2', '#bdd393', '#e56ffe', '#deff74', '#00ff78', 
            '#009bff', '#006401', '#0076ff', '#85a900', '#00b917', 
            '#788231', '#00ffc6', '#ff6e41', '#e85ebe'
            ]
    if rgb:
        rgbColors = []
        for color in colors:
            rgbColors.append(hex2rgb(color))
        colors = rgbColors

    if ignore:
        if isinstance(ignore, str):
            ignore = [ignore]
        for ig in ignore:
            try:
                colors.pop(colors.index(ig))
            except ValueError:
                pass

    return colors

def tardir(dir_, rm = True):
    
    if not os.path.isdir(format_path(dir_)):
        return False
    with tarfile.open(format_path(dir_)[:-1] + '.tar.gz', 'w:gz') as tar:
        tar.add(dir_, arcname = os.path.basename(format_path(dir_)[:-1]))
    if rm:
        shutil.rmtree(dir_)

def untardir(dir_, rm = False, to = None):
    if not to:
        to = os.path.dirname(dir_[:-1])
    tar = tarfile.TarFile.open(dir_)
    tar.extractall(path = to)
    tar.close()
    if rm:
        os.remove(dir_)

def checkdir(dir_, unzip = False, to = None, rm = False):
    if os.path.isdir(dir_):
        return True
    elif os.path.isfile(format_path(dir_) + '.tar.gz'):
        if unzip:
            if dir_.endswith('/'):
                dir_ = dir_[:-1]
            untardir(dir_ + '.tar.gz', rm = rm, to = to)
            return True
    return False


def eprint( *args, **kwargs ):
    '''Prints to stderr'''

    print(*args, file = sys.stderr, **kwargs)


def fprint(out_str, log):
    with open(log, 'a') as out:
        out.write(args)

def zprint(out_str, log = None, flush = True):
    fprint(out_str, log)
    print(out_str, flush = flush)

def vprint( toPrint, v = False, e = False , flush = True):
    '''Boolean print option to stdout or stderr (e)'''

    if v:
        if e:
            eprint( toPrint, flush = True)
        else:
            print( toPrint, flush = True)


def read_json( config_path, compress = False ):

    if compress or config_path.endswith('.gz'):
        with gzip.open( config_path, 'rt' ) as json_raw:
            json_dict = json.load( json_raw )
    else:
        with open( config_path, 'r' ) as json_raw:
            json_dict = json.load( json_raw )

    return json_dict
   

def write_json(obj, json_path, compress = False, indent = 1, **kwargs):
    if compress or json_path.endswith('.gz'):
        with gzip.open(json_path, 'wt') as json_out:
            json.dump(obj, json_out, indent = indent, **kwargs)
    else:
        with open(json_path, 'w') as json_out:
            json.dump(obj, json_out, indent = indent, **kwargs)


def gunzip( gzip_file, remove = True, spacer = '\t' ):
    '''gunzips gzip_file and removes if successful'''

    new_file = re.sub( r'\.gz$', '', format_path(gzip_file) )
    try:
        with gzip.open(format_path(gzip_file), 'rt') as f_in:
            with open(new_file, 'w') as f_out:
                for line in f_in:
                    f_out.write(line)
        if remove:
            os.remove( gzip_file )
        return new_file
    except:
        if os.path.isfile( new_file ):
            if os.path.isfile( gzip_file ):
                os.remove( new_file )
        raise IOError('gunzip ' + str(gzip_file) + ' failed')


def fmt_float(val, sig_dig = None):

    val_str = str(val)
    try:
        e = val_str.index('e')
        dig = val_str[:e].replace('.','')
        num = val_str[e + 1:]
        if num.startswith('-'):
            val_op = '0.'
            for i in range(abs(int(num))-1):
                val_op += '0'
            val_str = val_op + dig.replace('.','')
        else:
            num = abs(int(num))
            if len(dig) > num:
                val_op = list(dig) 
                val_op.insert('.',num+1)
                val_str = ''.join(val_op)
            else:
                zeroes = num + 1 - len(dig)
                val_op = dig
                for i in range(zeroes):
                    val_op += '0'
                val_str = val_op
    except ValueError:
        val_str = str(val)

    if sig_dig:
        if len(val_str.replace('.','')) > sig_dig:
            try:
                per_i = val_str.index('.')
            except ValueError:
                per_i = None
            val_list = list(val_str)
            if per_i:
                if sig_dig > per_i:
                    i = val_list[sig_dig]
                    post = int(val_list[sig_dig+1])
                    if post >= 5:
                        val_list.insert(sig_dig, int(i) + 1)
                        val_list.pop(sig_dig + 1)
                    val_str = ''.join(str(v) for v in val_list[:sig_dig + 1])
                else:
                    i = val_list[sig_dig - 1]
                    try:
                        post = val_list[sig_dig]
                    except IndexError:
                        post = 0
                    if post == '.':
                        post = val_list[sig_dig + 1]
                    elif i == '.':
                        i = val_list[sig_dig]
                        try:
                            post = val_list[sig_dig + 1]
                        except IndexError:
                            post = 0
                    if int(post) >= 5:
                        val_list.insert(sig_dig, int(i) + 1)
                        val_list.pop(sig_dig + 1)
                    val_list = val_list[:sig_dig]
                    while len(val_str) > len(val_list):
                        val_list.append('0')
                    val_str = ''.join(str(v) for v in val_list[:sig_dig+3])
            else:
                i = val_list[sig_dig + 1]
                post = int(val_list[sig_dig + 2])
                if post == '.':
                    post = int(val_list[sig_dig + 3])
                elif i == '.':
                    i = val_list[sig_dig + 2]
                    try:
                        post = val_list[sig_dig + 3]
                    except IndexError:
                        post = 0
                if int(post) >= 5:
                    val_list.insert(sig_dig + 1, int(i) + 1)
                    val_list.pop(sig_dig + 2)
                val_list = val_list[:sig_dig + 2]
                while len(val_str) > len(val_list):
                    val_list.append('0')
                val_str = ''.join(str(v) for v in val_list[:sig_dig + 1])
                

    return val_str




def findExecs( deps, exit = set(), verbose = True ):
    '''
    Inputs list of dependencies, `dep`, to check path.
    If dependency is in exit and dependency is not in path,
    then exit.
    '''

    vprint('\nDependency check:', v = verbose, e = True, flush = True)
    checks = []
    if type(deps) is str:
        deps = [deps]
    for dep in deps:
        check = shutil.which( dep )
        vprint('{:<15}'.format(dep + ':', flush = True) + \
            str(check), v = verbose, e = True)
        if not check and dep in exit:
            eprint('\nERROR: ' + dep + ' not in PATH', flush = True)
            sys.exit(300)
        else:
            checks.append(check)

    return checks


def findEnvs( envs, exit = set(), verbose = True ):
    '''
    Inputs list of paths, `envs`, to check path.
    If env is not in path and it is in exit, exit.
    '''

    vprint('\nEnvironment check:', v = verbose, e = True, flush = True)
    if type(envs) is str:
        envs = [envs]
    eprint(flush = True)
    for env in envs:
        try:
            vprint('{:<15}'.format(env + ':', flush = True) + \
                str(os.environ[env]), v = verbose, e = True)
        except KeyError:
            vprint('{:<15}'.format(env + ':', flush = True) + \
                'None', v = verbose, e = True)
            if env in exit:
                eprint('\nERROR: ' + env + ' not in PATH', flush = True)
                sys.exit(301)


def expandEnvVar( path ):
    '''Expands environment variables by regex substitution'''

    envs = re.findall(r'\$[^/]+', path)
    for env in envs:
        path = path.replace(env, os.environ[env.replace('$','')])

    return path.replace('//','/')


def format_path(path):
    '''Goal is to convert all path types to absolute path with explicit dirs'''

#    path = path.replace('//','/') 
#    except AttributeError: # not a string
 #       return None # removed this because let it be handled on the other end
 #   try: 
    if path:
        path = os.path.expanduser( path )
        path = expandEnvVar( path )
    #    path = os.path.abspath( path )
    #    except TypeError:
      #      return None again, let this be handled on the other end to increase
      #      throughput
    
        if path.endswith('/'):
            if not os.path.isdir( path ):
                path = path[:-1]
        else:
            if os.path.isdir( path ):
                path += '/'
        if not path.startswith('/'):
            path = os.getcwd() + '/' + path
    
    return path


# need to change recursive to false
def collect_files( directory = './', filetype = '*', recursive = False ):
    '''
    Inputs: directory path, file extension (no "."), recursivity bool
    Outputs: list of files with `filetype`
    If the filetype is a list, split it, else make a list of the one entry.
    Parse the environment variable if applicable. Then, obtain a clean, full 
    version of the input directory. Glob to obtain the filelist for each 
    filetype based on whether or not it is recursive.
    '''

    if type(filetype) == list:
        filetypes = filetype.split()
    else:
        filetypes = [filetype]

    directory = format_path( directory )
    filelist = []
    for filetype in filetypes:
        if recursive:
            filelist.extend( 
                glob.glob( directory + "/**/*." + filetype, recursive = recursive )
            )
        else:
            filelist.extend(
                glob.glob( directory + "/*." + filetype, recursive = recursive ) 
            )

    return filelist


def collect_folders( directory, recursive = False ):
    '''
    Inputs: directory path, recursive search boolean
    Outputs: list of folders
    Get folders via os, recursively extend output list if True.
    '''

    directory = format_path( directory )
    folders_prep = os.listdir( directory )
    folders = [ 
        directory + '/' + folder + '/' for folder in folders_prep \
        if os.path.isdir( directory + '/' + folder ) 
    ]

    if recursive:
        for folder in folders:
            folders.extend(collect_folders( folder, recursive ))

    return folders


def dictSplit( Dict, factor ):
    '''
    Inputs: a dictionary `Dict`, and an integer `factor` to split by
    Outputs: a list of split dictionaries `list_dict`
    Split dictionary keys into a numpy array via the factor. For each
    key, append a new dictionary to the output list and populate it 
    with the list of keys within that factor. The last entry will be
    whatever amount is leftover.
    '''
    

    list_dict = []
    keys_list = np.array_split( list(Dict.keys()), factor )

    try:    
        for keys in keys_list:
            list_dict.append( {} )
            for key in keys:
                list_dict[-1][key] = Dict[key]
    except TypeError:
        list_dict = []
        for keys in keys_list:
            list_dict.append( {} )
            for key in keys:
                list_dict[-1][tuple(key)] = Dict[tuple(key)]

    return list_dict


def sys_start( args, usage, min_len, dirs = [], files = [] ):
    """args is a sys.argv typically, or a list of arguments.
    usage is the usage statement without formatting.
    min_len is the minimum number of arguments that should exist.
    dirs is a list of items that should be directories
    files is a list of items that should be files"""

    if '-h' in args or '--help' in args:
        print( '\n' + usage + '\n' , flush = True)
        sys.exit( 1 )
    elif len( args ) < min_len:
        print( '\n' + usage + '\n' , flush = True)
        sys.exit( 2 )
    elif not all( os.path.isfile( format_path(x) ) for x in files ):
        print( '\n' + usage , flush = True)
        eprint( 'ERROR: input file(s) do not exist\n' , flush = True)
        sys.exit( 3 )
    elif not all( os.path.isfile( format_path(x) ) for x in dirs ):
        print( '\n' + usage , flush = True)
        eprint( 'ERROR: input directory does not exist\n' , flush = True)
        sys.exit( 4 )

    return args

 
def intro( script_name, args_dict, credit='', log = False, stdout = True):
    '''
    Inputs: script_name string, args_dict dictionary of arguments, 
    credit string bool / path for output log
    Outputs: prints an introduction, returns start_time in YYYYmmdd format
    Creates a string to populate and format for the introduction using 
    keys as the left-most descriptor and arguments (values) as the right
    most. Optionally outputs a log according to `log` path.
    '''

    start_time = datetime.now()
    date = start_time.strftime( '%Y%m%d' )

    out_str = '\n' + script_name + '\n' + credit + \
        '\nExecution began: ' + str(start_time)

    for arg in args_dict:
        out_str += '\n' + '{:<30}'.format(arg.upper() + ':') + \
            str(args_dict[ arg ])

    if log:
        zprint(out_str, log)
    elif stdout:
        print(out_str, flush = True)
    else:
        eprint( out_str, flush = True )

    return start_time


def outro( start_time, log = False, stdout = True ):
    '''
    Inputs: start time string formatted YYYYmmdd, log path
    Outputs: prints execution time and exits with 0 status
    '''

    
    end_time = datetime.now()
    duration = end_time - start_time
    dur_min = duration.seconds/60
    out_str = '\nExecution finished: ' + str(end_time) + '\t' + \
            '\n\t{:.2}'.format(dur_min) + ' minutes\n'

    if log:
        zprint(out_str, log)
    elif not stdout:
        eprint(out_str, flush = True)
    else:
        print(out_str, flush = True)

    sys.exit(0)


def prep_output(output, mkdir = True, require_newdir = False, cd = False):
    '''
    Inputs: output path, bool `mkdir`, bool `require_newdir`, bool cd.
    Outputs: creates directory, returns None if bool inputs fail, returns 
    formatted path
    '''

    output = format_path( output )
    if os.path.isdir( output ):
        if require_newdir:
            eprint('\nERROR: directory exists.', flush = True)
            return None
    elif os.path.exists( output ):
        output = os.path.dirname(output)
    else:
        if not mkdir:
            eprint('\nERROR: directory does not exist.', flush = True)
            return None
        os.mkdir(output)

    if cd:
        os.chdir(output)

    return output

def mkOutput(base_dir, program, reuse = True, suffix = datetime.now().strftime('%Y%m%d')):
    if not os.path.isdir(format_path(base_dir)):
        raise FileNotFoundError(base_dir + ' does not exist')
    out_dir = format_path(base_dir) + program + '_' + suffix

    if not reuse:
        count, count_dir = 1, out_dir
        while os.path.isdir(count_dir):
            count_dir  += '_' + str(count)
            count += 1
        os.mkdir(count_dir)
        return count_dir + '/'
    else:
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)
        return out_dir + '/'


def checkDep( dep_list = [], var_list = [], exempt = set() ):
    '''Checks all dependencies in path from list, optional exemption set'''

    failedVars = []
    check = [ shutil.which( dep ) for dep in dep_list if dep not in exempt ]
    for var in var_list:
        if var not in exempt:
            try:
                os.environ[ var ]
                check.append( True )
            except KeyError:
                check.append( False )
                failedVars.append( var )
    if not all( check ):
        eprint('\nERROR: Dependencies not met:', flush = True)
        for dep in dep_list:
            if not shutil.which( dep ):
                eprint( dep + ' not in PATH' , flush = True)
        for failed in failedVars:
            eprint( failed + ' variable not set' , flush = True)
        sys.exit( 135 )


def file2list(file_, types = '', sep = None, col = None, compress = False):
    '''
    Inputs: `file_` path, separator string, column integer index
    Outputs: reads file and returns a list given arguments
    If the `sep` is not valid, return None, else read the file and
    split each line, if `col`, split each by the delimiter, and
    acquire the column index at each line. If not `col`, then 
    simply split by the separator and grab the only entry.
    '''

    if not compress:
        with open(file_, 'r') as raw_data:
            data = raw_data.read() + '\n'
    else:
        with gzip.open(file_, 'rt') as raw_data:
            data = raw_data.read() + '\n'

    if col:
        check = data.split('\n')
        if sep:
            check_col = check[0].split(sep).index( col )
            data1_list = [ 
                x[check_col].rstrip() \
                for x in [y.split(sep) for y in check[1:]] \
                if len(x) >= check_col + 1 
            ]
        else:
            check_col = check[0].split().index(col)
            data1_list = [ 
                x[check_col].rstrip() \
                for x in [y.split() for y in check[1:]] \
                if len(x) >= check_col + 1 
            ]
    elif types:
        if sep:
            data_list = data.split(sep = sep)
        else:
            data_list = data.split()
        if types == 'int':
            try:
                data1_list = [int(x.rstrip()) for x in data_list if x]
            except ValueError:
                data1_list = [int(float(x.rstrip())) for x in data_list if x]
        elif types == 'float':
            data_list = [float(x.rstrip()) for x in data_list if x]
    else:
        if sep:
            data_list = data.split( sep = sep )
        else:
            data_list = data.split()
        data1_list = [ x.rstrip() for x in data_list if x ]

    return data1_list


def multisub(args_lists, processes = 1, shell = False, 
             verbose = False):
    '''
    Inputs: list of arguments, integer of processes, subprocess `shell` bool
    Outputs: launches and monitors subprocesses and returns list of exit 
    information
    If processes are too few, opt to default. While there are entries in the
    `args_lists`: 
        while there are less entries in `running` than max processes
    and arguments remaining in the `args_lists`: Popen the first argument in 
    the list, append the subprocess information and command to `running` and
    delete the argument from `args_lists`. 
        If there are still args in the args_lists, for each index in running
        grab the handle, poll if it is done, and if it is done output the 
        return code and stdin information to outputs and delete the index
        from `running`.
        IF there are not arguments in args_lists, wait until all processes
        are complete before exiting the function.
    '''

    if verbose:
        stdout, stderr = None, None
    else:
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL

    running, outputs = [], []
    if processes <= 0:
        processes = 1

    while args_lists: # while there are arguments populating arg_lists
        while len(running) < processes and args_lists: # while not @ max
        # processes
            cmd = args_lists[0]
            run_temp = subprocess.Popen(cmd , stdout = stdout, \
                stderr = stderr, shell = shell) # run command
            running.append([run_temp, cmd]) # add command to running
            del args_lists[0] # delete from argument list to run

        if len(args_lists) > 0: # if there are remaining arguments
            for index, handle in enumerate(running): # check each status
                handle[0].poll()
                returncode = handle[0].returncode
                if returncode is not None:
                    outputs.append({
                        'stdin': handle[1], 'code': returncode
                        }) # add command and exit code
                    del running[index] # remove from running
                    break

        else: # if there aren't remaining arguments
            while len(running) > 0: # hold until all commands complete
                handle = running[0]
                handle[0].poll()
                returncode = handle[0].returncode
                if returncode is not None:
                    outputs.append({
                        'stdin': handle[1], 'code': returncode
                        })
                    del running[0]

    return outputs # [{stdin: '', code: int()}]



#! /usr/bin/env python3

# NEED to convert gff list to appropriate types

import re
import sys
from mycotools.lib.kontools import eprint

aa_weights = {
    'A': 89.1, 'R': 174.2, 'N': 132.1, 'D': 133.1, 'C': 121.2, 'E': 147.1,
    'Q': 146.2, 'G': 75.1, 'H': 155.2, 'I': 131.2, 'L': 131.2, 'K': 146.2,
    'M': 149.2, 'F': 165.2, 'P': 115.1, 'S': 105.1, 'T': 119.1, 'W': 204.2,
    'Y': 181.2, 'V': 117.1
    }
aa_weights['X'] = sum(aa_weights.values())/len(aa_weights)

# codon translation table
translation_table = {
    'GCT':'A', 'GCC':'A', 'GCA':'A', 'GCG':'A',
    'CGT':'R', 'CGC':'R', 'CGA':'R', 'CGG':'R', 'AGA':'R', 'AGG':'R',
    'AAT':'N', 'AAC':'N',
    'GAT':'D', 'GAC':'D',
    'TGT':'C', 'TGC':'C',
    'CAA':'Q', 'CAG':'Q',
    'GAA':'E', 'GAG':'E',
    'GGT':'G', 'GGC':'G', 'GGA':'G', 'GGG':'G',
    'CAT':'H', 'CAC':'H',
    'ATT':'I', 'ATC':'I', 'ATA':'I',
    'TTA':'L', 'TTG':'L', 'CTT':'L', 'CTC':'L', 'CTA':'L', 'CTG':'L',
    'AAA':'K', 'AAG':'K',
    'ATG':'M',
    'TTT':'F', 'TTC':'F',
    'CCT':'P', 'CCC':'P', 'CCA':'P', 'CCG':'P',
    'TCT':'S', 'TCC':'S', 'TCA':'S', 'TCG':'S', 'AGT':'S', 'AGC':'S',
    'ACT':'T', 'ACC':'T', 'ACA':'T', 'ACG':'T',
    'TGG':'W',
    'TAT':'Y', 'TAC':'Y',
    'GTT':'V', 'GTC':'V', 'GTA':'V', 'GTG':'V',
    'TAA':'*', 'TGA':'*', 'TAG':'*'
}

rev_comp_table = {
    'T': 'A', 'G': 'C', 'C': 'G', 'A': 'T',
    'N': 'N', 'Y': 'R', 'R': 'Y', 'K': 'M',
    'M': 'K', 'B': 'V', 'D': 'H', 'H': 'D',
    'V': 'B', 't': 'a', 'g': 'c', 'c': 'g',
    'a': 't', 'n': 'n', 'y': 'r', 'r': 'y',
    'k': 'm', 'm': 'k', 'b': 'v', 'd': 'h',
    'h': 'd', 'v': 'b', 'W': 'W', 'S': 'S',
    'w': 'w', 's': 's'
    }


def calc_weight(seq):
    seq = seq.replace('*','')
    weight = aa_weights[seq[0]]
    for char in seq[1:]:
        weight += aa_weights[char] - 18.01
    return weight



def reverse_complement(seq):

    new_seq = ''
    for i in seq[::-1]:
        new_seq += rev_comp_table[i]

    return new_seq


def fa2dict(fasta_input, file_ = True):
    fasta_dict = {}
    if file_:
        with open(fasta_input, 'r') as fasta:
            str_fasta = ''
            for line in fasta:
                data = line.rstrip()
                if data.startswith('>'):
                    str_fasta += '\n' + data.rstrip() + '\n'
                elif data:
                    str_fasta += data
                # concatenates the prep string with the prepped line from the fasta file
        str_fasta = str_fasta.lstrip()
    else:
        str_fasta = fasta_input

    # extracts 0) seq ID and description and 1) sequence
    extracted = re.findall( r'(^>[^\n]*)\n([^>]*)', str_fasta, re.M)

    # adds a new dictionary for each gene
    for val in extracted:
        step1 = val[0]
        step2 = re.search(r'^>([^ ]*)', step1)
        gene = step2[1]
        step3 = re.search(r' (.*)', step1)
        if step3:
            descrip = step3[1]
        else:
            descrip = ''

        seq = val[1]
        if seq[-1] == '\n' or seq[-1] == '\r':
            seq = seq.rstrip()

        # prepares dictionaries for each gene with description, seq, rvcmpl_seq, and codons
        fasta_dict[gene] = {}
        if descrip != '\n':
            fasta_dict[gene]['description'] = descrip
        fasta_dict[gene]['sequence'] = seq

    return fasta_dict


# truncates sequences based on inputted lenght
def dnatrunc(fasta_dict,trunc_length):
    for gene in fasta_dict:
        if len(fasta_dict[gene]['sequence']) >= trunc_length:
            fasta_dict[gene]['sequence'] = fasta_dict[gene]['sequence'][:(trunc_length - 1)]
            # I should be able to do this in one line
            fasta_dict[gene]['reverse_complement'] = fasta_dict[gene]['reverse_complement'][-1:(0-trunc_length):-1]
            fasta_dict[gene]['reverse_complement'] = fasta_dict[gene]['reverse_complement'][::-1]
    return(fasta_dict)

# takes output dict from fasta2Dict and extracts codons from all 6 reading frames
def dict2codon(fasta_dict):

    fasta_dict[gene]['codons'] = {}

    # outputs a dictionary of reading frames and the possible codons for each
    for index in range(3):
        fasta_dict[gene]['codons']['reading_frame_' + str(index)] = []
        codon = ''
        for nt in fasta_dict[gene]['sequence'][index:]:
            if len(codon) < 3:
                codon = codon + nt
            if len(codon) == 3:
                fasta_dict[gene]['codons']['reading_frame_' + str(index)].append(codon)
                codon = ''

    # let's extract codons from all reading frames for the reverse sequences too
    for index in range(3):
        fasta_dict[gene]['codons']['reverse_reading_frame_' + str(index)] = []
        codon = ''
        for nt in fasta_dict[gene]['reverse_complement'][index:]:
            if len(codon) < 3:
                codon = codon + nt
            if len(codon) == 3:
                fasta_dict[gene]['codons']['reverse_reading_frame_' + str(index)].append(codon)
                codon = ''

    return codondict


def dict2fa(fasta_dict, description = True):

    fasta_string = ''
    if description:
        for gene in fasta_dict:
            fasta_string += '>' + gene.rstrip()
            if 'description' in fasta_dict[gene]:
                fasta_string += (' ' + fasta_dict[gene]['description']).rstrip() + '\n'
            else:
                fasta_string += '\n'
            fasta_string += fasta_dict[gene]['sequence'].rstrip() + '\n'

    else:
        for gene in fasta_dict:
            fasta_string += '>' + gene.rstrip() + '\n' + \
            fasta_dict[ gene ][ 'sequence' ].rstrip() + '\n'

    return fasta_string.rstrip()

# need to turn these into classes and class functions
def calc_gc(gene):

    G = gene['sequence'].count('G')
    C = gene['sequence'].count('C')
    GC = (G + C)/len(gene['sequence'])
    GC_con = '{:.2%}'.format(GC)

    return(GC_con)



# need to change into a class
def gff2list(gff_info, path = True, error = True):

    gff_list_dict = []
    if path:
        with open( gff_info, 'r' ) as raw_gff:
            data = [x.split('\t') for x in raw_gff.read().split('\n') \
                    if x and not x.startswith('#')]
    else:
        data = [x.split('\t') for x in gff_info.split('\n') \
                if x and not x.startswith('#')]
    try:
        for col_list in data:
            gff_list_dict.append({
                'seqid': col_list[0], 'source': col_list[1], 'type': col_list[2],
                'start': int(col_list[3]), 'end': int(col_list[4]), 'score': col_list[5],
                'strand': col_list[6], 'phase': col_list[7],
                'attributes': col_list[8]
                })
    except IndexError:
        raise IndexError(str(len(col_list)) + '/9 expected tab-' \
                        + 'delimitted fields: ' + str(line))
    except ValueError:
        raise ValueError(str(col_list[4:6]) + ' invalid integer ' \
                        + 'conversion: ' + str(line))
            
    return gff_list_dict


def list2gff(gff_list, ver = 3):

    if ver:
        gff_str = '##gff-version ' + str(ver) + '\n'
    else:
        gff_str = ''
    for line in gff_list:
        add_str = '\t'.join(str(val) for val in list(line.values()))
        gff_str += add_str + '\n'

    return gff_str.rstrip()

def gff3Comps( source = None ):

    comps = {}
    comps['par'] = r'Parent=["\']?([^;\'"]+)'
    comps['id'] = r'ID=["\']?([^;\'"]+)'
    comps['Alias'] = r'Alias=["\']?([^;\'"]+)'
    comps['product'] = r"""product=["\']?([^;"']+)"""
    comps['OG'] = r'OG=["\']?([\w+:\d+\|]+)'
    comps['ver'] = 'gff3'
    comps['prot'] = r'protein_id=["\']?([^;\'"]+)|proteinId=["\']?([^"\';]+)'
    comps['transcript'] = r'transcriptId=["\']?([^;"\'])'
   
    return comps

def gff2Comps():

    comps = {}
    comps['id'] = r'name "([^"]+)"'
    comps['prot'] = r'proteinId ([^;]+)'
    comps['transcript'] = r'transcriptId ([^;]+)'
    comps['alias'] = r'alias "([^"]+)"'
    comps['product'] = r'product_name "([^"]+)'
    comps['ver'] = 'gff2'

    return comps

def gtfComps():

    comps = {}
    comps['id'] = r'gene_id "([^"]+)"'
    comps['transcript'] = r'transcript_id "([^"]+)"'
    comps['alias'] = r'alias "([^"]+)"'
    comps['ver'] = 'gtf'

    return comps

def compileExon( gff ):
	
    exon_dict = {}

    for index in range(len(gff)):
        if gff[index]['type'].lower() == 'exon':
            break

    protComp = re.compile( r';Parent\=([^;]*)' )
    if not protComp.search( gff[index]['attributes'] ):
        protComp = re.compile( r'gene_id "(.*?)"' )
        if not protComp.search( gff[index]['attributes'] ):
            protComp = re.compile( r'name "(.*?)"\;' )
            if not protComp.search( gff[index]['attributes'] ):
                protComp = re.compile( r'ID=(.*?);' )

    for line in gff:
        if line['type'].lower() == 'exon':
            prot = protComp.search( line['attributes'] )[1]
            if prot not in exon_dict:
                exon_dict[ prot ] = []
            if line['strand'] == '+':
                exon_dict[prot].append( [int(line['start'] ) - 1, int( line['end'] )] )
            else:
                if int(line['start']) > int(line['end']):
                    exon_dict[prot].append( [int (line['end'] ) - 1, int( line['start'])])
                else:
                    exon_dict[prot].append( [int (line['start'] ) - 1, int( line['end'])])

    for prot in exon_dict:
        exon_dict[ prot ] = sorted(exon_dict[ prot ], key = lambda i: i[0])

    return exon_dict



#! /usr/bin/env python3

# NEED to try MCL using the binary
# NEED to make outgroup detection within the bounds of maximum sequences, optional to do so
    # NEEDs heavy optimization in general
# extract a fasta of the cluster of interest
# fix output
# NEED to avoid rerunning already ran values, only when necessary
# NEED to archive old runs'
# NEED to implement linclust

import os
import re
import sys
import copy
import shutil
import string
import random
import tempfile
import argparse
import itertools
import subprocess
import pandas as pd
from collections import defaultdict
from mycotools.lib.kontools import multisub, findExecs, format_path, \
    eprint, vprint, read_json, write_json, mkOutput, fmt_float
from mycotools.lib.biotools import fa2dict, dict2fa
sys.setrecursionlimit(1000000)

class ClusteringError(Exception):
    pass

class RefineError(Exception):
    pass

def run_mmseqs(fa_path, res_base, wrk_dir, algorithm = 'mmseqs easy-linclust',
                 min_id = 0.3, min_cov = 0.5, cpus = 1, verbose = False):
    res_path = res_base + '_cluster.tsv'
    tmp_dir = os.path.dirname(res_path) + '/tmp/'
    if verbose:
        stdout, stderr = None, None
    else:
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL
    mmseqs_cmd = algorithm.split(' ')
    mmseqs_cmd.extend([fa_path, res_base, tmp_dir,
                       '--min-seq-id', fmt_float(min_id), '--threads',
                       str(cpus), '--compressed', '1',
                       '--cov-mode', '0', '-c', str(min_cov),
                       '-e', '0.1'])
    mmseqs_exit = subprocess.call(mmseqs_cmd,
                                  stdout = stdout,
                                  stderr = stderr)
    if mmseqs_exit:
        raise ClusteringError('Clustering failed: ' + str(mmseqs_exit) \
                            + ' ' + str(mmseqs_cmd))
    if os.path.isfile(res_base + '_all_seqs.fasta'):
        os.remove(res_base + '_all_seqs.fasta')
    if os.path.isfile(res_base + '_rep_seq.fasta'):
        os.remove(res_base + '_rep_seq.fasta')
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    return res_path

def parse_mmseqs_clus(res_path):
    derivations = defaultdict(list)
    with open(res_path, 'r') as raw:
        for line in raw:
            k, v = line.rstrip().split() # all white space
            derivations[k].append(v)

    hg_list = []
    for k, v in derivations.items():
        v.append(k)
        hg_list.append(sorted(set(v)))

    hg2gene = {i: v for i, v in enumerate(
                    sorted(hg_list, key = lambda x: len(x),
                    reverse = True)
                    )}
    gene2hg = {}
    for hg, genes in hg2gene.items():
        for gene in genes:
            gene2hg[gene] = hg
    return hg2gene, gene2hg

def makeDmndDB(diamond, queryFile, output_dir, cpus = 1):
    """create a diamond database:
    diamond: binary path, queryFile: query_path"""
    outputDB = output_dir + re.sub(r'\.[^\.]+$', '', os.path.basename(queryFile))
    cmd = [
        diamond, 'makedb', '--in', queryFile,
        '--db', outputDB, '--threads', str(cpus)
        ]
    dmndDBcode = subprocess.call(
        cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
        )
    return outputDB, dmndDBcode

def runDmnd(
    diamond, queryFile, queryDB, outputFile, pid = True,
    cpus = 1, verbose = False, blast = 'blastp', minid = 15
    ):
    """run diamond, no self-hits, more-sensitive"""

    if pid:
        distanceVal = 'pident'
    else:
        distanceVal = 'bitscore'

    cmd = [
        diamond, blast, '-d', queryDB, '-q', queryFile,
        '--out', outputFile, '--threads', str(cpus),
        '--outfmt', '6', 'qseqid', 'sseqid', distanceVal,
        '--id', str(minid), '--no-self-hits', '--more-sensitive'
        ]

    if not verbose:
        dmndCode = subprocess.call(
            cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
            )
    else:
        dmndCode = subprocess.call(
            cmd
            )
    return outputFile, dmndCode

def rd_dmnd_distmtx( outputFile, minVal, pid = True ):
    """Read a diamond output as a distance matrix"""

    dist_dict = {}
    with open(outputFile, 'r') as raw:
        if pid:
            for line in raw:
                q, s, v0 = line.rstrip().split('\t')
                v = round(float(v0))
                if q == s:
                    continue
                if v > minVal:
                    if q not in dist_dict:
                        dist_dict[q] = {}
                        dist_dict[q][q] = 0.0
                    if s not in dist_dict:
                        dist_dict[s] = {}
                        dist_dict[s][s] = 0.0
                    dist_dict[q][s] = 100 - v
                    dist_dict[s][q] = 100 - v
            distance_matrix = pd.DataFrame(dist_dict).sort_index(1).sort_index(0).fillna(100)
        else:
             for line in raw:
                q, s, v0 = line.rstrip().split('\t')
                v = float(v0)
                if q == s:
                    continue
                if v > minVal:
                    if q not in dist_dict:
                        dist_dict[q] = {}
                        dist_dict[q][q] = 0.0
                    if s not in dist_dict:
                        dist_dict[s] = {}
                        dist_dict[s][s] = 0.0
             distance_matrix = pd.DataFrame(dist_dict).sort_index(1).sort_index(0).fillna(0)


    if pid:
        distance_matrix /= 100
    else:
        minV, maxV = min(distance_matrix), max(distance_matrix)
        denom = maxV - minV
        distance_matrix -= minV
        distance_matrix /= denom
        distance_matrix = 1 - distance_matrix

    return distance_matrix #, outMatrix

def runUsearch( fasta, output, clus_parameter, cpus = 1, verbose = False ):

    if verbose:
        subprocess.call([
            'usearch', '-calc_distmx', fasta,
            '-tabbedout', output, '-clus_parameter',
            clus_parameter, '-threads', str(cpus)
            ] #stdout = subprocess.PIPE,
            #stderr = subprocess.PIPE
            )
    else:
        subprocess.call([
            'usearch', '-calc_distmx', fasta,
            '-tabbedout', output, '-clus_parameter',
            clus_parameter, '-threads', str(cpus)
            ], stdout = subprocess.DEVNULL,
            stderr = subprocess.DEVNULL
            )

def rd_usrch_distmtx( dis_path, sep = '\t' ):
    '''Imports a distance matrix with each line formatted as `organism $SEP organism $SEP distance`.
    - this is equivalent to the `-tabbedout` argument in `usearch -calc_distmx`. The function 
    compiles a dictionary with the information and reciprocal information for each organism in each
    line, then converts this dictionary of dictionaries into a pandas dataframe. As a distance matrix,
    NA values are converted to maximum distance (1).'''

    distance_matrix = pd.DataFrame()
    with open( dis_path, 'r' ) as raw:
        data = raw.read()
    prepData = [ x.rstrip() for x in data.split('\n') if x != '' ]

    dist_dict = {}
    for line in prepData:
        vals = line.split( sep )
        if vals[0] not in dist_dict:
            dist_dict[ vals[0] ] = {}
        if vals[1] not in dist_dict:
            dist_dict[ vals[1] ] = {}
        dist_dict[ vals[0] ][ vals[1] ] = float(vals[2])
        dist_dict[ vals[1] ][ vals[0] ] = float(vals[2])

    distance_matrix = pd.DataFrame( dist_dict ).sort_index(0).sort_index(1)

    return distance_matrix.fillna( 1.0 )

def scikitaggd( distance_matrix, maxDist = 0.6, linkage = 'single' ):
    '''Performs agglomerative clustering and extracts the cluster labels, then sorts according to
    cluster number. In the future, this will also extract a Newick tree.'''

    clustering = AgglomerativeClustering( 
        affinity = 'precomputed',
        distance_threshold = maxDist,
        n_clusters = None,
        linkage = linkage
        ).fit( distance_matrix )

    i = iter( distance_matrix.columns )
    clusters = { next( i ): x for x in clustering.labels_ }
    clusters = { k: v for k, v in sorted( clusters.items(), key = lambda item: item[1] ) }

    return clusters

def getNewick(node, newick, parentdist, leaf_names):
    '''Code from https://stackoverflow.com/questions/28222179/save-dendrogram-to-newick-format
    adopted from @jfn'''
    if node.is_leaf():
        return "%s:%.2f%s" % (leaf_names[node.id], parentdist - node.dist, newick)
    else:
        if len(newick) > 0:
            newick = "):%.2f%s" % (parentdist - node.dist, newick)
        else:
            newick = ");"
        newick = getNewick(node.get_left(), newick, node.dist, leaf_names)
        newick = getNewick(node.get_right(), ",%s" % (newick), node.dist, leaf_names)
        newick = "(%s" % (newick)
        return newick

def getClusterLabels( labels, clusters ):

    i = iter(labels)
    protoclusters = {next(i): x for x in clusters}
    clusters = {k: v for k, v in sorted(protoclusters.items(), key = lambda item: item[1])}

    return clusters

def runMCL(distMat, inflation):

    key2gene = {i: v for i, v in enumerate(list(distMat.keys()))}
    npMat = distMat.to_numpy()
    result = mc.run_mcl(npMat, inflation = inflation)
    protoClusters = mc.get_clusters(result)

    clusters = {}
    for i, keys in enumerate(protoClusters):
        clusters = {**clusters, **{key2gene[key]: i for key in keys}}

    return clusters

def scipyaggd( distMat, maxDist, method = 'single' ):
    '''Performs agglomerative clustering using SciPy and extracts the cluster labels, then sorts
    according to cluster number.'''

    squareform_matrix = squareform(distMat.values)
    linkage_matrix = hierarchy.linkage(squareform_matrix, method)
    tree = hierarchy.to_tree(linkage_matrix)
    fcluster = hierarchy.fcluster(
        linkage_matrix, maxDist, 
        criterion = 'distance'
        )
    clusters = getClusterLabels(distMat.index, fcluster)

    return clusters, tree

def dmnd_main(fasta_path, minVal, output_dir, distFile, pid = True, verbose = False, cpus = 1):
    queryDB, makeDBcode = makeDmndDB('diamond', fasta_path, output_dir, cpus = cpus)
    dmndOut, dmndCode = runDmnd(
        'diamond', fasta_path, queryDB, distFile + '.tmp', pid = pid,
        cpus = cpus, verbose = verbose, blast = 'blastp'
        )
    shutil.move(distFile + '.tmp', distFile)
    distance_matrix = rd_dmnd_distmtx( distFile, minVal, pid = pid )
    return distance_matrix
    
def usrch_main(fasta, min_id, output, cpus = 1, verbose = False):
    vprint('\nusearch aligning', flush = True, v = verbose)
    runUsearch(fasta, output + '.dist', str(1 - min_id), cpus, verbose)
    distance_matrix = rd_usrch_distmtx( output + '.dist' )
    return distance_matrix

def readLog(log_path, newLog):
    oldLog = read_json(log_path)
    if oldLog['fasta'] == newLog['fasta'] and \
        oldLog['minimum_id'] == newLog['minimum_id'] and \
        oldLog['search_program'] == newLog['search_program']:
        newLog['iterations'] = oldLog['iterations']
    elif newLog['distance_matrix']:
        if os.path.isfile(newLog['distance_matrix']):
            os.remove(newLog['distance_matrix'])
    return newLog

def run_iterative(
    minseq, maxseq, clus_parameter, mincon, params, focal_gene, 
    interval = 0.1, log_dict = None, log_path = None, minval = 0, 
    maxval = 1, verbose = False, spacer = '\t', cpus = 1
    ):
    """
    Iteratively clusters until input parameters are met.

    minseq: int(minimum sequences in a cluster)
    maxseq: int(maximum sequences in a cluster)
    clus_parameter: float(maximum_distance for agglomerative), float(minimum
        coverage for linclust)
    mincon: minimum connection value
    distance_matrix: distance_matrix generated from fa2clus functions
    focal_gene: str(gene to cluster around)
    linkage: agglomerative clustering linkage parameter
    interval: interval to adjust clus_parameter per iteration
    log_dict: log
    log_path: log_file path
    minval: int(minimum clus_parameter value)
    maxval: int(maximum clus_parameter value)
    verbose: bool(verbosity)
    spacer: str(stderr/stdout spacer)
    """

    exit_code = 1 # default exit code is failure
    attempt, direction, clusters, tree = 0, None, None, None
    oldDirection, oldClusters, oldTree = None, None, None
    while clus_parameter >= minval and clus_parameter <= maxval:
        attempt += 1 # keep track of iterations for output details
        if 'dist' in params: # hierarchical clustering
            if 1 - clus_parameter < mincon:
                break
            oldClusters, oldTree = clusters, tree
            clusters, tree = scipyaggd(params['distance_matrix'],
                                       float(clus_parameter), 
                                       params['linkage']) # cluster
            cluster_dict = defaultdict(list) # could be more efficient by grabbing in getClusterLabels
            for gene, index in clusters.items(): # create a dictionary clusID:
            # [genes]
                cluster_dict[index].append(gene)
        else:
            res_base = params['dir'] + focal_gene
            res_path = run_mmseqs(params['fa'], res_base, params['dir'], 
                                    algorithm = params['bin'],
                                    min_id = clus_parameter, verbose = verbose,
                                    min_cov = params['min_cov'], cpus = cpus)
            cluster_dict, clusters = parse_mmseqs_clus(res_path)
        focalLen = len(cluster_dict[clusters[focal_gene]]) 
        # size of cluster with focal_gene

        vprint('\nITERATION ' + str(attempt) + ': ' + focal_gene \
             + ' cluster size: ' + str(focalLen), flush = True, v = verbose)
        vprint('Cluster parameter: ' + str(clus_parameter), flush = True, v = verbose)
        if log_path: # output the iteration details to the log
            iteration_dict = {'size': focalLen, 'cluster_parameter': clus_parameter}
            log_dict['iterations'].append(iteration_dict)
            write_json(log_dict, log_path)
        if focalLen >= minseq: # if greater than minimum sequences
            if maxseq: # if there is a max set of sequences
                if focalLen <= maxseq: # if less than max sequences
                    vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                    exit_code = 0
                    break
                else: # descend
                    direction = 1
            else: # no max sequences, minimum is met, this was successful
                vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                exit_code = 0
                break
        else: # need to increase cluster size
            direction = -1

        if oldDirection: # if there is an older iteration
            if direction != oldDirection: # if the direction switched
                eprint(spacer + 'WARNING: Overshot - ' \
                     + 'could not find parameters using current interval', 
                       flush = True)
                vprint(spacer + 'Outputting Iterations ' + str(attempt) \
                     + ' & ' + str(attempt - 1), flush = True, v = verbose)
                exit_code = 2
                break
        else: # set the old iteration direction detail
            oldDirection = direction
#         mincon += (interval*direction)

        oclus_parameter = copy.copy(clus_parameter)
        if 'dist' in params:
            clus_parameter -= (interval*direction) 
        else:
            clus_parameter += (interval*direction)

        # clus_parameter should be to 3 significant figures
        clus_parameter = round(clus_parameter * 100)/100
        # adjust the clus_parametereter based on the direction it should change
        if oclus_parameter != minval and oclus_parameter != maxval:
            if clus_parameter < minval: # set as minimum value if it descends below
                clus_parameter = minval
            elif clus_parameter > maxval: # set as maximum if it ascends above
                clus_parameter = maxval

    return clusters, tree, oldClusters, oldTree, log_dict, exit_code

def run_reiterative(
    minseq, maxseq, clus_parameter, mincon, params, focal_gene, interval = 0.01,
    log_dict = None, log_path = None, minval = 0, maxval = 1, verbose = False,
    cpus = 1
    ):
    """Goal is to refine cluster size as close to maxseq as much as possible"""

    attempt, direction, clusters, tree = 0, None, None, None
    oldDirection, refines = None, []
    while clus_parameter >= minval and clus_parameter <= maxval: # while a valid cluster
    # parameter
        attempt += 1 # log attempts
        if 'dist' in params: # hierarchical clustering
            oldClusters, oldTree = clusters, tree
            clusters, tree = scipyaggd(params['distance_matrix'],
                                       float(clus_parameter), 
                                       params['linkage']) # cluster
            cluster_dict = defaultdict(list) # could be more efficient by grabbing in getClusterLabels
            for gene, index in clusters.items(): # create a dictionary clusID:
            # [genes]
                cluster_dict[index].append(gene)
            focalLen = len(cluster_dict[clusters[focal_gene]])
            vprint('\nITERATION ' + str(attempt) + ': ' + focal_gene \
                 + ' cluster size: ' + str(focalLen), flush = True, v = verbose)
            vprint('\tMinimum connection: ' + str(mincon) \
                 + '; Maximum cluster parameter' + str(clus_parameter), flush = True, v = verbose)
        else: # mmseqs
            res_base = params['dir'] + focal_gene
            res_path = run_mmseqs(params['fa'], res_base, params['dir'], 
                                    algorithm = params['bin'],
                                    min_id = clus_parameter, verbose = verbose,
                                    min_cov = params['min_cov'], cpus = cpus)
            cluster_dict, clusters = parse_mmseqs_clus(res_path)
            focalLen = len(cluster_dict[clusters[focal_gene]])
            vprint('\nITERATION ' + str(attempt) + ': ' + focal_gene \
                 + ' cluster size: ' + str(focalLen), flush = True, v = verbose)
            vprint('\tMinimum coverage: ' + str(mincon) \
                 + '; Minimum identity: ' + str(clus_parameter), flush = True, v = verbose)
        # find the length of the cluster with the gene
        if log_path: # output the log
            iteration_dict = {'size': focalLen, 'cluster_parameter': clus_parameter}
            log_dict['iterations'].append(iteration_dict)
            write_json(log_dict, log_path)
        if focalLen >= minseq: # if the focal cluster is > minimum size
            if maxseq: # check if there is a maximum size
                if focalLen <= maxseq: # if less than the maximum
                    refines.append([clusters, tree, True]) # append the info to
                    # refine info
                    direction = -1 # we can climb
                else:
                    refines.append([clusters, tree, False]) # else we need to
                    # go down
                    direction = 1
            else:
                # if there is no maximum size, then we met the goal
                return clusters, tree, None, None, log_dict
        else: # if less than the minimum size, then we need to climb
            refines.append([clusters, tree, False])
            direction = -1

        if oldDirection:
            if direction != oldDirection: # if the directions changed it was
            # met/overshot
                if refines[-2][2]: # if the second to last run was successful
                    return refines[-2][0], refines[-2][1], None, None, log_dict
                else: # return the two closest to success
                    return refines[-1][0], refines[-1][1], refines[-2][0], refines[-2][1], log_dict
        else:
            oldDirection = direction

        oclus_parameter = copy.copy(clus_parameter)
        if 'dist' in params: # if hierachical agglomerative clustering
            clus_parameter -= (interval*direction)
        else:
            clus_parameter += (interval*direction)

        # clus_parameter should be to 3 significant figures
        clus_parameter = round(clus_parameter * 100)/100

        if oclus_parameter != minval and oclus_parameter != maxval:
            if clus_parameter < minval:
                clus_parameter = minval
            elif clus_parameter > maxval:
                clus_parameter = maxval

    return refines[-1][0], refines[-1][1], None, None, log_dict


def main(
    fasta_path, mincon, clus_parameter, minseq, maxseq, 
    search_program = 'diamond', linkage = 'single',
    iterative = False, interval = 0.1, output= None, cpus = 1,
    verbose = False, spacer = '\t\t', log_path = None,
    refine = False, pid = True, dmnd_dir = None
    ):

    if search_program in {'usearch', 'diamond'}:
        algorithm = 'hierarchical'
        dist_path = output + '.dist'
    else:
        algorithm = search_program
        dist_path = None

    minval, maxval = 0, 1
    overshot = False
    log_dict = {
        'algorithm': algorithm, 'fasta': fasta_path, 
        'minimum_id': mincon, 'cluster_parameter': clus_parameter,
        'minimum_sequences': minseq, 'maximum_sequences': maxseq, 
        'search_program': search_program, 'distance_matrix': dist_path, 
        'linkage': linkage, 'focal_gene': iterative,
        'iterations': []
        }
    if log_path:
        if os.path.isfile(log_path):
            log_dict = readLog(log_path, log_dict)
        write_json(log_dict, log_path)

    if algorithm == 'hierarchical':
        global hierarchy
        global squareform
        from scipy.cluster import hierarchy
        from scipy.spatial.distance import squareform
        param_dict = {'dist': None, 'link': linkage}
        if os.path.isfile(log_dict['distance_matrix']):
            if search_program == 'usearch':
                param_dict['dist'] = rd_usrch_distmtx(log_dict['distance_matrix'])
            else:
                param_dict['dist'] = rd_dmnd_distmtx(log_dict['distance_matrix'], mincon, pid = pid)
        else:
            if search_program == 'diamond': #elif if above lines not highlighted
                if not dmnd_dir:
                    dmnd_dir = os.path.dirname(log_dict['distance_matrix']) + '/'
                param_dict['dist'] = dmnd_main(
                    fasta_path, mincon, dmnd_dir, 
                    log_dict['distance_matrix'], pid = pid, 
                    verbose = verbose, cpus = cpus
                    )
            elif search_program == 'usearch':
                param_dict['dist'] = usrch_main(fasta_path, mincon, output, cpus, verbose)
    else:
        param_dict = {'fa': fasta_path, 'dir': os.path.dirname(output) + '/',
                      'bin': search_program, 'min_cov': mincon}

    vprint('\nClustering', flush = True, v = verbose)
    if iterative:
        clusters, tree, oldClusters, oldTree, log_dict, exit_code = run_iterative(
            minseq, maxseq, clus_parameter, mincon, param_dict,
            iterative, interval = interval, 
            log_dict = log_dict, log_path = log_path, minval = minval, maxval = maxval,
            verbose = verbose, spacer = spacer, cpus = cpus
            )
        if refine:
            if exit_code == 0 or exit_code == 2: # parameters were met/overshot
                if len(log_dict['iterations']) > 1: # if there's room to work
                    vprint('\nMaximizing cluster size', v = verbose, flush = True)
                    beyond_iters = [x for x in log_dict['iterations'] \
                                    if int(x['size']) > maxseq]
                                    # iterations with too many sequences
                    sorted_beyonds = sorted(beyond_iters, key = lambda x: \
                                            (x['size'], x['cluster_parameter']))
                                            # sort smallest to largest
                    beyonds = [x for x in sorted_beyonds \
                               if x['size'] == sorted_beyonds[0]['size']]
                               # grab all iterations of the fewest of too
                               # many
                    final_iteration = log_dict['iterations'][-1]
                    if algorithm == 'hierarchical':
                        minval = final_iteration['cluster_parameter']
                        new_cp = minval
                        if beyonds:
                            interval = 0.005
                            maxval = float(beyonds[0]['cluster_parameter'])
                        else:
                            interval = 0.05
                            if pid:
                                maxval = 1 - mincon # max is relative to % ID cutoff
                            else:
                                maxval = 1
                    else: # mmseqs
                        if beyonds:
                            interval = 0.01
                            minval = float(beyonds[-1]['cluster_parameter'])
                            # smallest value surpassing max sequences
                        else:
                            interval = 0.05
                            minval = 0.0
                        maxval = final_iteration['cluster_parameter']
                        # the working value
                        if minval > maxval:
                            raise RefineError('Cannot refine mmseqs cluster')
                        new_cp = maxval
                    clusters, tree, oldClusters, oldTree, log_dict = run_reiterative(
                        minseq, maxseq, new_cp, mincon, param_dict, iterative, 
                        interval = interval, log_dict = log_dict, 
                        log_path = log_path, minval = minval,
                        maxval = maxval, verbose = verbose, cpus = cpus
                        )
                    if oldTree:
                        vprint(
                            spacer + 'WARNING: Cluster does not exist within parameters', 
                            v = verbose, flush = True
                            )
            else: # parameters were not met
                vprint(
                    spacer + 'WARNING: Could not find a cluster that meets parameters', 
                    v = verbose, flush = True
                    )
        elif exit_code == 0:
            vprint(spacer + 'SUCCESS!', v = verbose, flush = True)
        elif exit_code == 1:
            vprint(
                spacer + 'WARNING: Could not find a cluster that meets parameters', 
                v = verbose, flush = True
                )
        else:
            vprint(
                spacer + 'WARNING: Overshot parameters',
                v = verbose, flush = True
                )
            overshot = True
    else:
        oldClusters, oldTree = None, None
        if algorithm == 'hierarchical':
            clusters, tree = scipyaggd(distance_matrix, clus_parameter, linkage)
        else:
            res_base = output
            res_path = run_mmseqs(fasta_path, output, 
                                    os.path.dirname(output) + '/', 
                                    verbose = verbose,
                                    algorithm = algorithm, min_id = clus_parameter, 
                                    min_cov = mincon, cpus = cpus)
            null_dict, clusters = parse_mmseqs_clus(res_path)

    if algorithm == 'hierarchical':
        return distance_matrix, clusters, tree, oldClusters, oldTree, overshot
    else:
        return None, clusters, None, oldClusters, None, overshot

def write_data(tree, distance_matrix, clusters, output):

    #clusters = scikitaggd( distance_matrix, float(args.max_dist), args.linkage )

    out_str = '\n'.join([i + '\t' + str(clusters[i]) for i in clusters])
    with open(output + '.clus', 'w') as out:
        out.write( out_str )
    if tree:
        newick = getNewick(tree, "", tree.dist, list(distance_matrix.index))
        with open(output + '.newick', 'w') as out:
            out.write(newick)


if __name__ == '__main__':

    parser = argparse.ArgumentParser( 
        description = "Sequence clustering with iterative search option." 
        )
    parser.add_argument('-f', '--fasta', required = True)
    parser.add_argument( 
        '-a', '--alignment', 
        help = 'Alignment software {"mmseqs", "diamond", "usearch"}; ' \
             + 'diamond and usearch call hierarchical clustering. ' \
             + 'DEFAULT: mmseqs', 
        default = 'mmseqs' 
        )
    parser.add_argument('-l', '--linclust', action = 'store_true',
        help = 'Use linclust instead of mmseqs cluster (faster, less sensitive)')
    parser.add_argument('-m', '--mincon', help = 'Minimum connection' \
        + ' for alignment; DEFAULT: [mmseqs]: 0.1 percent query coverage, '  \
        + '[aggclus]: 0.2 percent identity, [aggclus]: 30 bitscore', 
        type = float)
    parser.add_argument('-x', '--cluster_parameter', 
        help = '[mmseqs]: Minimum identity; DEFAULT: 0.3; ' \
            +  '[aggclus]: Maximum distance; DEFAULT: 0.75', 
            type = float)
    parser.add_argument('-i', '--iterative', 
        help = '[--minseq]: Gene to iteratively cluster for; '  \
            + 'eliminates -x')
    parser.add_argument('--minseq', type = int, help = '[-i]: Iteratively adjust -m and -x until a cluster ' \
        + ' of minimum sequences is obtained; DEFAULT: 2', default = 2)
    parser.add_argument('--maxseq', type = int, 
        help = '[-i]: Maximum sequences for iterative clustering; ' \
            + 'DEFAULT 100', default = 100)
    parser.add_argument('-r', '--refine', action = 'store_true', 
        help = '[--maxseq]: Refine cluster to maximize sequences within parameters.')
#    parser.add_argument('-i', '--inflation', type = float, default = 1.5,
 #       help = 'Inflation value for MCL; DEFAULT: 1.5')
    parser.add_argument(
        '-d', '--distance_type', help = '[aggclus]: {"identity", "bitscore"} ' + \
        'bitscore is normalized 0-1 following -m and -x should be 0-1; DEFAULT: identity',
        default = 'identity'
        )

    parser.add_argument('--linkage', default = 'single', 
        help = "[aggclus]: Linkage criterion: " \
        + "'complete' (maximum distance), 'average', 'weighted', " \
        + "'centroid', or 'single' (minimum distance) DEFAULT: 'single' (minimum)")
    parser.add_argument('-o', '--output')
    parser.add_argument('-c', '--cpus', default = 1, type = int)
    parser.add_argument('-v', '--verbose', action = 'store_true')
    args = parser.parse_args()

    if args.refine and not args.maxseq:
        eprint('\nERROR: --maxseq required for refinement', flush = True)
    if args.alignment in {'diamond', 'usearch'}:
        if args.linkage not in { 'complete', 'average', 'weighted', 'centroid', 'single' }:
            eprint('\nERROR: Invalid linkage criterium', flush = True)
            sys.exit( 1 )
        elif args.distance_type not in {'identity', 'bitscore'}:
            eprint('\nERROR: Invalid distance type', flush = True)
            sys.exit( 3 )
        if args.distance_type == 'identity':
            pid = True
            if not args.mincon:
                args.mincon = 0.2
        else:
            pid = False
            if not args.mincon:
                args.mincon = 30
        if not args.cluster_parameter:
            args.cluster_parameter = 0.75
    elif args.algnment == 'mmseqs':
        if args.linclust:
            args.alignment = 'mmseqs easy-linclust'
        else:
            args.alignment = 'mmseqs easy-cluster'
        if not args.mincon:
            args.mincon = 0.1
            pid = None
        if not args.cluster_parameter:
            args.cluster_parameter = 0.3
    else:
        eprint('\nERROR: Invalid alignment software', flush = True)
        sys.exit( 2 )

    findExecs([args.alignment], exit = set(args.alignment))
    interval = 0.1

    if args.iterative and args.alignment in {'diamond', 'usearch'}:
        clus_parameter = 1 - args.mincon
    elif 1 - args.cluster_parameter <= args.mincon \
        and args.alignment != 'mmseqs':
        eprint('\nWARNING: 1 - maximum distance exceeds minimum connection, clustering is ineffective', flush = True)
    else:
        clus_parameter = args.cluster_parameter
 #   else:
#        clus_parameter = args.inflation

    fasta_path = format_path(args.fasta)
    fa = fa2dict(fasta_path)
    if args.iterative:
        focal_gene = args.iterative
        if len(fa) < args.minseq:
            eprint('\nERROR: minimum sequences is greater than fasta input', flush = True)
            sys.exit(5)
        elif focal_gene not in fa:
            eprint('\nERROR: ' + focal_gene + ' not in ' + fasta_path, flush = True)
            sys.exit(6)

    if args.output:
        if not os.path.isdir(format_path(args.output)):
            os.mkdir(format_path(args.output))
        dmnd_dir = format_path(args.output)
        output = dmnd_dir + os.path.basename(fasta_path)
    else:
        dmnd_dir = mkOutput(os.getcwd() + '/', 'fa2clus')
        output = dmnd_dir + os.path.basename(fasta_path)

    distance_matrix, clusters, tree, ocl, ot, overshot = main(
        fasta_path, args.mincon, clus_parameter, args.minseq, args.maxseq, 
        search_program = args.alignment, linkage = args.linkage, verbose = args.verbose,
        iterative = args.iterative, interval = interval, output = output,
        log_path = dmnd_dir + '.' + os.path.basename(fasta_path) + '.fa2clus.json',
        refine = args.refine, pid = pid, dmnd_dir = dmnd_dir, cpus = args.cpus,
        spacer = '\n'
        )

    if ocl:
        write_data(tree, distance_matrix, clusters, output + '.0')
        write_data(ot, distance_matrix, ocl, output + '.1')
    else:
        write_data(tree, distance_matrix, clusters, output)

    if args.iterative:
        cluster_dict = {}
        for gene, index in clusters.items():
            if index not in cluster_dict:
                cluster_dict[index] = []
            cluster_dict[index].append(gene)
    
        gene_module = cluster_dict[clusters[focal_gene]]
        input_fa = fa2dict(fasta_path)
        output_fa = {x: input_fa[x] for x in gene_module}
   
        with open(output + '.fa', 'w') as out:
            out.write(dict2fa(output_fa)) 

    sys.exit(0)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
from shutil import copy as cp
from mycotools.lib.dbtools import masterDB, mtdb
from mycotools.lib.kontools import format_path, prep_output, eprint, vprint


def soft_main(filetypes, db, output_path, print_link = False, verbose = False):
    """Symlink or print files from each file_type"""

    db = db.set_index('ome')
    if not print_link:
        for ftype in filetypes:
            if not os.path.isdir(output_path + ftype):
                os.mkdir(output_path + ftype)
        for ome, row in db.items():
            for ftype in filetypes:
                if os.path.isfile(row[ftype]):
                    try:
                        os.symlink(row[ftype], output_path + ftype \
                                 + '/' + ome + '.' + ftype)
                    except FileExistsError:
                        vprint('\t' + ome + ' ' + ftype + ' exists',
                               v = verbose, flush = True)
                else:
                    vprint('\tERROR: ' + ome + ' ' + ftype, flush = True,
                           v = verbose)
    else:
        for ome, row in db.items():
            for ftype in filetypes:
                print(row[ftype], flush = True)


def hard_main(filetypes, db, output_path):
    """Hard copy files from filetypes to their filetype output directory"""

    db = db.set_index('ome')
    for ftype in filetypes:
        if not os.path.isdir(output_path + ftype):
            os.mkdir(output_path + ftype)

    for ome, row in db.items():
        for ftype in filetypes:
            try:
                cp(row[ftype], output_path + ftype + '/' \
                + os.path.basename(row[ftype])) 
            except FileNotFoundError:
                eprint('\tERROR: ' + ome + ' ' + ftype, flush = True)

if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = 'Symlinks/copies selected files from database' )
    parser.add_argument( '-d', '--mtdb', default = masterDB(), help = 'DEFAULT: masterDB' )
    parser.add_argument( '-a', '--assembly', action = 'store_true', help = 'Grab assemblies' )
    parser.add_argument( '-p', '--proteome', action = 'store_true', help = 'Grab proteomes' )
    parser.add_argument( '-g', '--gff', action = 'store_true', help = 'Grab gff`s' )
    parser.add_argument( '--print', action = 'store_true', help = 'Print paths, no copy' )
    parser.add_argument( '--hard', action = 'store_true', help = 'Hard copy files' )
    parser.add_argument( '-o', '--output', default = os.getcwd() )
    args = parser.parse_args()

    if not args.assembly and not args.proteome and not args.gff:
        print('\nERROR: no file type (-a, -g, -p) selected', flush = True)
        sys.exit(4)

    db_path = format_path(args.mtdb)
    output_path = prep_output(format_path( args.output ), cd = False)
    args_dict = {
        'DATABASE': db_path, 'OUTPUT': output_path,
        'ASSEMBLY': args.assembly, 'PROTEOME': args.proteome,
        'GFF3': args.gff, 'Print links': args.print, 'Hard copy': args.hard
        }

    filetypes = []
    if args.proteome:
        filetypes.append('faa')
    if args.gff:
        filetypes.append('gff3')
    if args.assembly:
        filetypes.append('fna')

    db = mtdb(db_path)
    if args.print or not args.hard:
        soft_main(filetypes, db, output_path, print_link = args.print)
    else:
        hard_main(filetypes, db, output_path)

    sys.exit(0)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
from collections import defaultdict
from mycotools.lib.biotools import fa2dict, dict2fa, reverse_complement
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.kontools import format_path, eprint, stdin2str

def extract_mtdb_accs_exp(fa_dict, accs):
    return {acc: fa_dict[acc] for acc in accs}

def extract_mtdb_accs(fa_dict, accs):
    """extract MTDB accessions from fa_dict using a list `accs`; can 
    extract coordinates of an accession if labeled 
    `acc[START:END] | acc[START-END]`. If start > end, extract the reverse
    complement of the sequence. Returns a fa_dict of the results"""

    out_fa = {}
    for acc in accs:
        coords = acc[acc.find('['):acc.find(']')]
        if coords: # if there are coordinates
            try: # try splitting by colon
                start, end = [int(x) for x in coords.split(':')]
            except ValueError: # try splitting by hyphen
                try:
                    start, end = [int(x) for x in coords.split('-')]
                except ValueError: # invalid coordinates, try whole acc
                    out_fa[acc] = fa_dict[acc]
                    continue
            acc_name = acc[:start]
            if start < end:
                out_fa[acc] = {
                    'sequence': fa_dict[acc_name]['sequence'][start:end+1],
                    'description': fa_dict[acc_name]['description']
                    }
            else:
                out_fa[acc] = {
                    'sequence': reverse_complement(
                        fa_dict[acc_name]['sequence'][end-1:start+1]
                        ),
                    'description': fa_dict[acc_name]['description']
                    }
        else: # no coordinates
            out_fa[acc] = fa_dict[acc] # extract the whole accession

    return out_fa
            

def extractHeaders(fasta_file, accessions, ome = None):
    """searches headers for "[]", which indicate coordinate-based extraction.
    otherwise, just retrieves the accession from the fasta dictionary"""

    fasta = fa2dict(fasta_file)
    out_fasta = {}
    acc_comp = re.compile( r'([^/[]*)\[(\d+)\-(\d+)\]$' )

    for header in accessions:
        if '[' in header: # if coordinate based
            coord_search = acc_comp.search( header )
            if coord_search is not None:
                start = int(coord_search[2]) - 1
                end = int(coord_search[3])
                acc = coord_search[1]
                if ome:
                    header = ome + '_' + header
                if end > start:
                    out_fasta[header] = {
                        'sequence': fasta[acc]['sequence'][start:end],
                        'description': ''
                        }
                else:
                    out_fasta[header] = {
                        'sequence': reverse_complement(
                            fasta[acc]['sequence'][end-1:start+1]
                            ),
                        'description': ''
                        }
            else:
                if ome:
                    out_fasta[ome + '_' + header] = fasta[header]
                else:
                    out_fasta[header] = fasta[header]
        else:
            if ome:
                out_fasta[ome + '_' + header] = fasta[header]
            else:
                out_fasta[header] = fasta[header]

    return out_fasta
   

def dbmain(db, accs, error = True, spacer = '\t\t\t', coord_check = True):
    """takes in mtdb, takes accessions ome by ome from accs;
    error will exit if set to true, report to stderr if not;
    coord_check will check entries for coordinate signatures in accession names
    but will bypass if false (throughput increase). Returns a fa_dict of the
    accessions and coordinates if applicable"""

    # set the db index
    db = db.set_index( 'ome' )
    # create a dict to populate for each ome
    ome_data = defaultdict(list)
    # organize accession by ome
    for acc in accs:
        ome = acc[:acc.find('_')]
        ome_data[ome].append(acc)

    # populate the fasta dict with the accession information by referencing the
    # mycotools proteome entry for each ome
    fa_dict = {}
    if coord_check:
        for ome, ome_accs in ome_data.items():
            try:
                ome_fasta = fa2dict(db[ome]['faa'])
            except KeyError: # if there is a missing ome
                if error:
                    raise KeyError
                else:
                    eprint(spacer + ome + ' not in database', flush = True)
            fa_dict = {**fa_dict, **extract_mtdb_accs(ome_fasta, ome_accs)}
    else:
        for ome, ome_accs in ome_data.items():
            try:
                ome_fasta = fa2dict(db[ome]['faa'])
            except KeyError: # if there is a missing ome
                if error:
                    raise KeyError
                else:
                    eprint(spacer + ome + ' not in database', flush = True)
            fa_dict = {**fa_dict, **extract_mtdb_accs_exp(ome_fasta, ome_accs)}
 

    return fa_dict


def famain( accs, fa, ome = None ):
    """takes in accessions, fasta, and retrieves accessions"""

    fa_dict = {}
    fa_dict = {**fa_dict, **extractHeaders(fa, accs, ome)}
        
    return fa_dict


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description = 'Inputs accession, extracts fasta')
    parser.add_argument('-a', '--accession', help = '"-" for stdin. For coordinates ' + \
        'append [$START-$END] - reverse coordinates for antisense')
    parser.add_argument('-i', '--input', help = 'File with accessions')
    parser.add_argument('-f', '--fasta', help = 'Fasta input')
    parser.add_argument('-c', '--column', default = 1, 
        help = 'Accessions column for -i (1 indexed). DEFAULT: 1', type = int)
    parser.add_argument('-s', '--start', help = 'Start index column (1 indexed)', type = int)
    parser.add_argument('-e', '--end', help = 'End index column (1 indexed)', type = int)
    parser.add_argument('-d', '--mtdb', default = masterDB())
    args = parser.parse_args()

    if args.input: # input file
        input_file = format_path(args.input)
        if args.start: # if using columns for coordinates
            with open(input_file, 'r') as raw:
                accs = []
                for line in raw:
                    if not line.startswith('#'):
                        d = line.rstrip().split('\t')
                        accs.append([
                            d[args.column-1] + '[' + d[args.start-1] + '-' + d[args.end-1] + ']'
                            ])
        else:
            with open(input_file, 'r') as raw:
                accs = [x.rstrip().split('\t')[args.column-1] for x in raw]
    else: # assume we are using accessions
        if '-' == args.accession: # stdin
            data = stdin2str()
            accs = data.split()
        else:
            accs = [args.accession]
       
    db_path = format_path(args.mtdb)
    if not args.fasta: # MTDB run
        db = mtdb(db_path)
        fa_dict = dbmain(db, accs)
        fasta_str = dict2fa(fa_dict)
    else: # non MTDB
        fa_path = format_path(args.fasta)
        fa_dict = famain(accs, fa_path)
        fasta_str = dict2fa(fa_dict)

    print(fasta_str.rstrip() , flush = True)
    sys.exit(0)



#!/usr/bin/env python3

import os
import re
import sys
import copy
import shutil
import multiprocessing as mp
from collections import Counter, defaultdict
from mycotools.lib.kontools import gunzip, mkOutput, format_path, eprint, vprint
from mycotools.lib.biotools import gff2list, list2gff, fa2dict, dict2fa, \
    gff3Comps, gff2Comps
from mycotools.lib.dbtools import mtdb, masterDB, loginCheck
from mycotools.curAnnotation import main as curRogue
from mycotools.utils.curGFF3 import main as curGFF3
from mycotools.utils.gff2gff3 import main as gff2gff3
from mycotools.gff2seq import aamain as gff2seq

predb_headers = [
    'assembly_accession', 'previous_ome', 
    'genus', 'species', 'strain', 'version', 'biosample',
    'assemblyPath', 'gffPath', 'genomeSource (ncbi/jgi/new)', 
    'useRestriction (yes/no)', 'published'
    ]

def prep_output(base_dir):
    out_dir = mkOutput(base_dir, 'predb2db')
    wrk_dir = out_dir + 'working/'
    dirs = [
        out_dir, wrk_dir, wrk_dir + 'gff3/', wrk_dir + 'fna/', wrk_dir + 'faa/'
        ]
    for dir_ in dirs:
        if not os.path.isdir(dir_):
            os.mkdir(dir_)
    return dirs[:2]

def copy_file( old_path, new_path ):

    try:
        shutil.copy( old_path, new_path )
        return True
    except:
        raise IOError


def move_biofile(old_path, ome, typ, wrk_dir, suffix = '' ):

    if old_path.endswith('.gz'):
        if not os.path.isfile(old_path[:-3]):
            temp_path = gunzip(old_path)
            new_path = wrk_dir + ome + '.' + typ +  suffix
        else:
            new_path = wrk_dir + ome + '.' + typ + suffix
            temp_path = old_path[:-3]
        copy_file(format_path(temp_path), new_path)
    else:
        new_path = wrk_dir + ome + '.' + typ +  suffix
        if not os.path.isfile(new_path) and os.path.isfile(old_path):
            copy_file(format_path(old_path), new_path)
        elif not os.path.isfile(new_path):
            raise IOError
        else:
            copy_file(format_path(old_path), new_path)

    return new_path


def gen_predb():
    example = [
        'Fibsp1', 'Fibularhizoctonia', 'psychrophila', 'CBS',
        '1.0', 'n', '<PATH/TO/ASSEMBLY>', '<PATH/TO/GFF3>', 'jgi',
        'no', '2018'
        ]
    eprint('INSTRUCTIONS: fill in each column with the relevant information and \
        separate each column by a tab. The predb can be filled in \
        via spreadsheet software and exported as a tab delimited `.tsv`. \
        ASSEMBLY ACCESSIONS and PREVIOUS_OME fields must be unique to the \
        genome; otherwise predb2db will update the corresponding database entry. \
        Novel data must be filled in as "new" for the genomeSource column.', flush = True)
    outputStr = '#' + '\t'.join(predb_headers)
    outputStr += '\n#' + '\t'.join(example) + '\n'

    return outputStr

def read_predb(predb_path, spacer = '\t'):

#    predb, headers = {}, None
    predb = defaultdict(list)
    with open(predb_path, 'r') as raw:
        for i, line in enumerate(raw):
#            if line.startswith('#'):
 #               if not headers:
  #                  predb = {x: [] for x in line.rstrip()[1:].split('\t')}
   #                 headers = list(predb.keys())
            if not line.startswith('#') and line.rstrip():
#            if line.rstrip():
                line = line.replace('\n','')
                entry = line.split('\t')
                if len(entry) != len(predb_headers):
                    eprint(spacer + 'ERROR: Incorrect columns, line ' + str(i),
                           flush = True)
                    eprint(predb_headers, '\n', entry, flush = True)
                    sys.exit(3)
                for i, v in enumerate(entry):
                    predb[predb_headers[i]].append(v.rstrip())

    predb = dict(predb)
    try:
        predb['assembly_acc'] = predb['assembly_accession']
        del predb['assembly_accession']
    except KeyError:
        if not 'assembly_acc' in predb and not 'assembly_accession' in predb:
            predb['assembly_acc'] = ['' for x in predb['genus']]
    try:
        predb['source'] = predb['genomeSource (ncbi/jgi/new)']
        del predb['genomeSource (ncbi/jgi/new)']
    except KeyError:
        if not 'genomeSource' in predb and not 'source' in predb:
            raise KeyError
    try:
        predb['restriction'] = predb['useRestriction (yes/no)']
        del predb['useRestriction (yes/no)']
        if any(x.lower() not in {'y', 'n', 'yes', 'no', '', 'true', 'false'} for x in predb['restriction']):
            eprint(spacer + 'ERROR: useRestriction entries must be in {y, n, yes, no}', flush = True)
            sys.exit(4)
    except KeyError:
        if not 'restriction' in predb and 'published' not in predb:
            raise KeyError
        elif 'published' in predb:
            predb['restriction'] = [bool(x) for x in predb['published']]

    if any(x.lower() not in {'jgi', 'ncbi', 'new'} \
        for x in predb['source']):
        eprint(spacer + 'ERROR: genomeSource entries must be in {jgi, ncbi, new}', flush = True)
        sys.exit(5)
    for i, path in enumerate(predb['assemblyPath']):
        predb['assemblyPath'][i] = format_path(predb['assemblyPath'][i])
        predb['gffPath'][i] = format_path(predb['gffPath'][i])
        if predb['restriction'][i].lower() in {'y', 'yes', 'true'}:
            predb['restriction'][i] = True
        elif predb['restriction'][i].lower() in {'n', 'no', 'false'}:
            predb['restriction'][i] = ''
        if not predb['species'][i]:
            predb['species'][i] = 'sp.'
    return predb

def predb2mtdb(predb):
    infdb = mtdb()
    if not 'assemblyPath' in predb and 'fna' in predb:
        predb['assemblyPath'] = predb['fna']
    if not 'gffPath' in predb and 'gff3' in predb:
        predb['gffPath'] = predb['gff3']
    if not 'previous_ome' in predb and 'ome' in predb:
        predb['previous_ome'] = predb['ome']
    elif not 'previous_ome' in predb:
        predb['previous_ome'] = \
            [None for x in predb['ome']]
    for i, code in enumerate(predb['genus']):
        if predb['published'][i]:
            toAdd = {
                'assembly_acc': predb['assembly_acc'][i],
                'ome': predb['previous_ome'][i],
                'genus': predb['genus'][i],
                'species': predb['species'][i],
                'strain': predb['strain'][i],
                'version': predb['version'][i],
                'biosample': predb['biosample'][i],
                'fna': predb['assemblyPath'][i],
                'gff3': predb['gffPath'][i],
                'source': predb['source'][i],
                'published': predb['published'][i]
                }
        else:
            toAdd = {
                'assembly_acc': predb['assembly_acc'][i],
                'ome': predb['previous_ome'][i],
                'genus': predb['genus'][i],
                'species': predb['species'][i],
                'strain': predb['strain'][i],
                'version': predb['version'][i],
                'biosample': predb['biosample'][i],
                'fna': predb['assemblyPath'][i],
                'gff3': predb['gffPath'][i],
                'source': predb['source'][i],
                'published': not predb['restriction'][i]
                }
#        if predb['published'][i]:
 #           toAdd['published'] = predb['published'][i]
        infdb = infdb.append(toAdd)
    return infdb

def gen_omes(
    newdb, refdb = None, ome_col = 'ome', forbidden = set(),
    spacer = '\t'
    ):

    t_failed = []
    ref_ome_check = [k for k,v in Counter(refdb['ome']).items() if v > 1]
    new_ome_check = [k for k,v in Counter(newdb['ome']).items() if v > 1 and k]
    if ref_ome_check:
        raise ValueError('corrupted reference database with non-unique omes: '
                        + str(ref_ome_check))
    elif new_ome_check:
        raise ValueError('corrupted pre-database with non-unique omes: '
                        + str(new_ome_check))
    tax_list = list(set(refdb['ome']).union(forbidden))
    tax_count = {}
    for tax in tax_list:
        abb = tax[:6]
        try:
            num = int(tax[6:])
        except ValueError: # version included in number
            num = int(re.search(r'(^\d+)', tax[6:])[1])
        if abb in tax_count:
            if tax_count[abb] < num:
                tax_count[abb] = num
        else:
            tax_count[abb] = num
    tax_count = Counter(tax_count)

    todel = []
    refdb_aas = refdb.set_index('assembly_acc')
    refdb_accs = set([x for x in list(refdb['assembly_acc']) if x])
    refdb_omes = set([x for x in list(refdb['ome'])])
    for i, ome in enumerate(newdb['ome']):
        if not ome: # if there isn't an ome for this entry yet
            if newdb['assembly_acc'][i] in refdb_accs: # if this is an
            # established assembly accession; maybe make sure this works for
            # changed MycoCosm or NCBI assembly accs
                ome = refdb_aas[newdb['assembly_acc'][i]]['ome']
                v_search = re.search(r'\.(\d+)$', ome[6:])
                if v_search:
                    v = int(v_search[1]) + 1 # new version
                    new_ome = re.sub(r'\.\d+$', '.' + str(v), ome)
                else:
                    new_ome = ome + '.1' # first modified version
                eprint(spacer + ome + ' update -> ' + new_ome, flush = True)
                newdb['ome'][i] = new_ome
                continue
               
            try:
                name = newdb['genus'][i][:3].lower() + newdb['species'][i][:3].lower()
            except TypeError:
                todel.append(i)
                if not isinstance(newdb['assembly_acc'][i], float):
                    eprint(spacer + newdb['assembly_acc'][i] + ' no metadata - ' \
                         + 'failed', flush = True)
                elif 'index' in newdb: # for updateDB
                    if not isinstance(newdb, float):
                        eprint(spacer + newdb['index'][i] + ' no metadata - ' \
                            + 'failed', flush = True)
                    continue 
                else: # no use appending failed when there's no identifiable
                # info
                    continue
                row = {key: newdb[key][i] for key in mtdb.columns \
                       if key != 'ome'}
                t_failed.append(add2failed(row))
                continue
            name = re.sub(r'\(|\)|\[|\]|\$|\#|\@| |\||\+|\=|\%|\^|\&|\*|\'|\"|\!|\~|\`|\,|\<|\>|\?|\;|\:|\\|\{|\}', '', name)
            name.replace('(', '')
            name.replace(')', '')
            while len(name) < 6:
                name += '.'
            tax_count[name] += 1
            new_ome = name + str(tax_count[name])
            newdb['ome'][i] = new_ome
        elif ome:
            format_search = re.search(r'^[^\d_,\'";:\\\|\[\]\{\}\=\+\!@#\$\%\^' \
                                    + r'&\*\(\)]{6}\d+[^-_+=\\\|\{\[\}\]\:;' \
                                    + r'\'\"\,\<\>\?/\`\~\!\@\#\$\%\^\&\*\(' \
                                    + r'\)\w\W]*$', ome) # crude format check
            if not format_search:
                raise ValueError('invalid ome ' + ome)
            if ome in refdb_omes: # it's an update
                v_search = re.search(r'\.(\d+)$', ome[6:])
                if v_search:
                    v = int(v_search[1]) + 1 # new version
                    new_ome = re.sub(r'\.\d+$', '.' + str(v), ome)
                else:
                    new_ome = ome + '.1' # first modified version
                eprint(spacer + ome + ' update -> ' + new_ome, flush = True)
                newdb['ome'][i] = new_ome

    for i in reversed(todel):
        for key in mtdb.columns:
            del newdb[key][i]

    return newdb, t_failed

def cur_fna(cur_raw_fna_path, uncur_raw_fna_path, ome):
    with open(cur_raw_fna_path + '.tmp', 'w') as out:
        with open(uncur_raw_fna_path, 'r') as in_:
            for line in in_:
                if line.startswith('>'):
                    if not line.startswith('>' + ome + '_'):
                        out.write('>' + ome + '_' + line[1:])
                    else: # already curated
                        out.write(line)
                else:
                    out.write(line)
    shutil.move(cur_raw_fna_path + '.tmp', cur_raw_fna_path)

def cur_mngr(ome, raw_fna_path, raw_gff_path, wrk_dir, 
            source, assembly_accession, exit = False,
            remove = False, spacer = '\t\t\t', verbose = False):

    predb_dir = os.path.basename(os.path.dirname(wrk_dir[:-1])) + '/working/'

    # assembly FNAs
    vprint('\t' + ome, v = verbose, flush = True)
    uncur_fna_path = wrk_dir + 'fna/' + ome + '.fna.uncur'
    cur_fna_path = wrk_dir + 'fna/' + ome + '.fna'
    vprint('\t\t' + predb_dir + 'fna/' + ome + '.fna', v = verbose, flush = True)
    if not os.path.isfile(cur_fna_path):
        try:
            uncur_fna_path = move_biofile(raw_fna_path, ome, 'fa', wrk_dir + 'fna/',
                                      suffix = '.uncur')
        except IOError:
            eprint(spacer + ome + '|' + assembly_accession \
                 + ' failed FNA parsing', flush = True)
            if exit:
                raise IOError
            return ome, False, 'fna'
        cur_fna(cur_fna_path, uncur_fna_path, ome)

    # gene coordinate GFF3s
    uncur_gff_path = wrk_dir + 'gff3/' + ome + '.gff3.uncur'
    cur_gff_path = wrk_dir + 'gff3/' + ome + '.gff3'
    vprint('\t\t' + predb_dir + 'gff3/' + ome + '.gff3', v = verbose, flush = True)
    if not os.path.isfile(cur_gff_path):
        try:
            uncur_gff_path = move_biofile(raw_gff_path, ome, 'gff3', 
                                          wrk_dir + 'gff3/', suffix = '.uncur')
        except IOError:
            eprint(spacer + ome + '|' + assembly_accession \
                 + ' failed GFF3 parsing', flush = True)
            return ome, False, 'gff3'
    
        gff = gff2list(uncur_gff_path)
        try:
            gff_mngr(ome, gff, cur_gff_path, source, assembly_accession)
        except Exception as e: # catch all errors to continue script
            eprint(spacer + ome + '|' + assembly_accession \
                + ' failed GFF3 curation', flush = True)
            if exit:
                raise e
            return ome, False, 'gff3'

    # proteome FAAs
    faa_path = wrk_dir + 'faa/' + ome + '.faa'
    vprint('\t\t' + predb_dir + 'faa/' + ome + '.faa', v = verbose, flush = True)
    if not os.path.isfile(faa_path):
        try:
            faa = gff2seq(gff2list(cur_gff_path), fa2dict(cur_fna_path),
                          spacer = spacer)
        except Exception as e: # catch all errors
            eprint(spacer + ome + '|' + assembly_accession \
                 + ' failed proteome generation', flush  = True)
            if exit:
                raise e
            return ome, False, 'faa'
        with open(faa_path + '.tmp', 'w') as out:
            out.write(dict2fa(faa))
        shutil.move(faa_path + '.tmp', faa_path)

    if remove:
        if os.path.isfile(uncur_gff_path):
            os.remove(uncur_gff_path)
        if os.path.isfile(raw_gff_path):
            os.remove(raw_gff_path)
        if os.path.isfile(re.sub(r'\.gz$', '', raw_gff_path)):
            os.remove(re.sub(r'\.gz$', '', raw_gff_path))
        if os.path.isfile(uncur_fna_path):
            os.remove(uncur_fna_path)
        if os.path.isfile(raw_fna_path):
            os.remove(raw_fna_path)
        if os.path.isfile(re.sub(r'\.gz$', '', raw_fna_path)):
            os.remove(re.sub(r'\.gz$', '', raw_fna_path))

    return ome, cur_fna_path, cur_gff_path, faa_path


def gff_mngr(ome, gff, cur_path, source, assembly_accession):

    gffVer, alias = None, False
    for entry in gff:
        if re.search(gff3Comps()['id'], entry['attributes']):
            gffVer = 3
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])
            break
        elif re.search(gff2Comps()['id'], entry['attributes']):
            gffVer = 2
            break

    if gffVer == 3:
        if source == 'new':
            if alias: #already curated
                aliasOme = alias[1]
                for entry in gff:
                    alias0 = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
                    alias_num = alias0[alias0.find('_') + 1:]
                    new_alias = ome + '_' + alias_num
                    entry['attributes'] = re.sub(
                        gff3Comps()['Alias'], 'Alias='+ new_alias,
                        entry['attributes']
                        )
            else:
                gff, trans_str, failed, flagged = curRogue(gff, ome,
                                                           cur_seqids = True)
        else:
            gff = curGFF3(gff, ome, cur_seqids = True)
    else:
        gff = gff2gff3(gff, ome, assembly_accession, verbose = False,
                       cur_seqids = True)

    with open(cur_path + '.tmp', 'w') as out:
        out.write(list2gff(gff))
    shutil.move(cur_path + '.tmp', cur_path)

def add2failed(row):
    if isinstance(row['assembly_acc'], float) or not row['assembly_acc']:
        return False
    else:
        return [row['assembly_acc'], row['version']]

def main(
    predb, refdb, wrk_dir, 
    verbose = False, spacer = '\t\t\t', forbidden = set(), 
    cpus = 1, exit = False, remove = False
    ):

    infdb = predb2mtdb(predb)
    vprint('\nGenerating omes', v = verbose, flush = True)
    omedb, failed = gen_omes(infdb, refdb, ome_col = 'ome', forbidden = forbidden,
                     spacer = spacer)
    
    cur_cmds = []
    omedb = omedb.set_index('ome')
    for ome, row in omedb.items():
        cur_cmds.append([
            ome, row['fna'], row['gff3'], 
            wrk_dir, row['source'], row['assembly_acc'],
            exit, remove, spacer, verbose
            ])

    vprint('\nCurating data', v = verbose, flush = True)
    with mp.Pool(processes = cpus) as pool:
        cur_data = pool.starmap(cur_mngr, cur_cmds)

    for data in cur_data:
        if not data[1]:
            failed.append(add2failed(omedb[data[0]]))
            del omedb[data[0]]
        else:
            ome, fna_path, gff3_path, faa_path = data
            omedb[ome]['fna'] = fna_path
            omedb[ome]['gff3'] = gff3_path
            omedb[ome]['faa'] = faa_path

    return omedb.reset_index(), failed


if __name__ == '__main__':
    usage = 'Generate a predb file:\npredb2db.py\n\nCreate a mycotoolsdb ' + \
    'from a predb file:\npredb2db.py <PREDBFILE>\n\nCreate a mycotoolsdb ' + \
    'referencing an alternative master database:\npredb2db.py <PREDBFILE> ' + \
    '<REFERENCEDB>\nSkip failing genomes:\npredb2db.py <PREDBFILE> -s'

    if any(x in {'-h', '--help', '-help'} for x in sys.argv):
        eprint('\n' + usage + '\n', flush = True)
        sys.exit(1)
    elif len(sys.argv) == 1:
        print(gen_predb())
        sys.exit(0)
    elif len(sys.argv) >= 3:
        if sys.argv[2] not in {'-s', '--skip'}:
            refDB = mtdb(format_path(sys.argv[2]))
        elif len(sys.argv) > 3:
            refDB = mtdb(format_path(sys.argv[3]))
        else:
            refDB = mtdb(masterDB())
    else:
        refDB = mtdb(masterDB())

    if set(sys.argv).intersection({'-s', '--skip'}):
        exit = False
    else:
        exit = True

    from Bio import Entrez
    ncbi_email, ncbi_api, jgi_email, jgi_pwd = loginCheck(jgi = False)
    Entrez.email = ncbi_email
    if ncbi_api:
        Entrez.api_key = ncbi_api

    eprint('\nPreparing run', flush = True)
    predb = read_predb(format_path(sys.argv[1]), spacer = '\t')
    out_dir, wrk_dir = prep_output(os.path.dirname(format_path(sys.argv[1])))


    omedb, failed = main(predb, refDB, wrk_dir, exit = exit, verbose = True)

    from mycotools.lib.dbtools import gather_taxonomy, assimilate_tax
    tax_dicts = gather_taxonomy(omedb, api_key = ncbi_api)
    outdb = assimilate_tax(omedb, tax_dicts)
    outdb.df2db(out_dir + 'predb2db.mtdb')
    sys.exit(0)



#! /usr/bin/env python3

# NEED to flag overlapping coordinates

import os
import sys
import argparse
from Bio.Seq import Seq
from collections import defaultdict
from mycotools.lib.biotools import fa2dict, dict2fa
from mycotools.lib.kontools import sys_start, eprint, format_path

def extractCoords(fa_dict, seqid, coord_start = 0, coord_end = -1, sense = '+', fa_name = ''):

    new_fa, error = {}, ''
    name = seqid + '_' + str(coord_start) + '-' + str(coord_end) + '_' + sense
    if sense == '+':
        new_fa[name] = {
            'sequence': fa_dict[seqid]['sequence'][coord_start:coord_end],
            'description': ''
            }
    else:
        new_fa[name] = {
            'sequence': str(Seq(fa_dict[seqid]['sequence'][coord_start:coord_end]).reverse_complement()),
            'description': ''
            }
    if coord_end > len(fa_dict[seqid]['sequence']):
        error = '\nNOTICE: end coordinate beyond contig edge: ' + fa_name + ' ' + seqid + ' ' + str(coord_end)

    return new_fa, error

if __name__ == '__main__':

    usage = 'Input nucleotide fasta/tsv input, extract coordinates\n' \
        + 'coords2fa.py <FA> <SEQID> <START_COORD> <END_COORD> <STRAND_SENSE>' \
        + '\nExtract full sequence from sense strand: coords2fa.py test.fna scaffold_20 0 -1' \
        + '\nExtract coordinates from antisense strand: coords2fa.py test.fna scaffold_20 69 420 -' \
        + '\n\n\nBulk extraction tab delimitted row format:' \
        + '\n#fasta_path\tsequence_id\tstart_coordinate\tend_coordinate\tstrand_sense\t[concat_id]' \
        +  '\n\ncoords2fa.py coords.tsv'

    args = sys_start(sys.argv[1:], usage, 1, files = [sys.argv[1]])

    out_fa, error = defaultdict(dict), ''
    try:
        fa_file = format_path(args[0])
        if fa_file.endswith('/'):
            fa_name = os.path.basename(fa_file[:-1])
        else:
            fa_name = os.path.basename(fa_file)

        fa = fa2dict(fa_file)
        if not fa:
            raise IndexError
        if len(args) < 3:
            args.extend([0, -1, '+'])
        elif len(args) < 6:
            args.append('+')
        out_fa, error = extractCoords(
            fa, args[1], min([int(args[2]), int(args[3])]), max([int(args[2]), int(args[3])]), args[4], fa_name
            )

        print(dict2fa(out_fa))
        eprint(error)
        sys.exit(0)

    except IndexError:
        try: 
            with open(args[0], 'r') as raw:
                data = [x.rstrip().split('\t') \
                        for x in raw if x and not x.startswith('#')]
            files_data = {} 
            for x in data:
                if x[0] not in files_data:
                    files_data[x[0]] = defaultdict(list)
                if len(x) > 5:
                    files_data[x[0]][int(x[5])].append((x[1], min([int(x[2]), 
                                            int(x[3])]), max([int(x[2]), 
                                            int(x[3])]), x[4],))
                else:
                    files_data[x[0]][None].append((x[1], min([int(x[2]), int(x[3])]),
                                            max([int(x[2]), int(x[3])]),
                                            x[4],))
                   
        except:
            eprint('\nERROR: incorrectly formatted input', flush = True)

    for fa_file, concats in files_data.items():
        fa = fa2dict(format_path(fa_file))
        if fa_file.endswith('/'):
            fa_name = os.path.basename(fa_file[:-1])
        else:
            fa_name = os.path.basename(fa_file)
        for concat_id, rows in concats.items():
            if rows[0][-1] == '+':
                sorted_rows = sorted(rows, key = lambda x: x[1])
            else:
                sorted_rows = sorted(rows, key = lambda x: x[1],
                                    reverse = True)
            if concat_id is None:
                for x in sorted_rows:
                    new_fa, error_t = extractCoords(fa, x[0], x[1], x[2], 
                                                    x[3], fa_name)
                    error += error_t
                    out_fa = {**out_fa, **{fa_name + '_' + x: new_fa[x] for x in new_fa}}
            else:
                if not all(x[-1] == rows[0][-1] for x in rows):
                    raise ValueError('conflicting strand sense for ' \
                                    + 'concatenation ID:', concat_id, fa_name)
                toadd_fa = {'description': '', 'sequence': ''}
                for x in sorted_rows:
                    new_fa, error_t = extractCoords(fa, x[0], x[1], x[2],
                                                    x[3], fa_name)
                    error += error_t
                    for seq, seq_info in new_fa.items():
                        toadd_fa['sequence'] += seq_info['sequence']
                out_fa = {**out_fa, 
                          **{fa_name + '_concat' + str(concat_id): toadd_fa}}

    print(dict2fa(out_fa))
    eprint(error)
    sys.exit(0)



#! /usr/bin/env python3

# NEED to implement multigene phylogeny
# NEED to implement fa2clus?

import os
import re
import sys
import argparse
import subprocess
from collections import defaultdict
from mycotools.lib.kontools import eprint, vprint, collect_files, \
    format_path, intro, outro, findExecs
from mycotools.lib.biotools import fa2dict

class PhyloError(Exception):
    pass

def mafftRun(fasta, out_dir, hpc, verbose = True, cpus = 1, spacer = '\t'):

    name = fasta
    if '/' in fasta:
        name = os.path.basename(os.path.abspath(fasta)) 
    cmd = [
        'mafft', '--auto', '--thread', '-' + str(cpus), fasta
        ]

    if not hpc:
        print(spacer + 'Aligning', flush = True)
        with open(out_dir + name + '.mafft', 'w') as out_file:
            if verbose:
                run_mafft = subprocess.call(cmd, stdout = out_file)
            else:
                run_mafft = subprocess.call(cmd, stdout = out_file, stderr = subprocess.DEVNULL)
         
        if run_mafft != 0:
            eprint(spacer + '\tERROR: mafft failed: ' + str(run_mafft), flush = True)
            if os.path.isfile(out_dir + name + '.mafft'):
                os.remove(out_dir + name + '.mafft')	
            raise PhyloError
    else:
        with open(out_dir + '/mafft.sh', 'w') as out:
            if '#PBS' in hpc:
                out.write( hpc + '\n\n' + ' '.join([str(x) for x in cmd]) + \
                ' > ' + out_dir + '/' + name + '.mafft' + \
                '\ncd ' + out_dir + '\nqsub clipkit.sh')
            else:
                 out.write( hpc + '\n\n' + ' '.join([str(x) for x in cmd]) + \
		' > ' + out_dir + '/' + name + '.mafft' + \
                '\ncd ' + out_dir + '\nsbatch clipkit.sh')
        
    return out_dir + '/' + name + '.mafft'


def trimRun(mafft, out_dir, hpc, verbose, cpus = 1, spacer = '\t'):

    name2 = os.path.basename(os.path.abspath(re.sub( r'\.mafft$', '.clipkit', mafft)))
    cmd = ['clipkit', mafft, '--output', out_dir + '/' + name2]
    if not hpc:

        print(spacer + 'Trimming', flush = True)
        if verbose:
            run_clipkit = subprocess.call(cmd, stdout = subprocess.PIPE)
        else:
            run_clipkit = subprocess.call( 
                cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
                )
        if run_clipkit != 0:
            eprint(spacer + '\tERROR: `clipkit` failed: ' + str(run_clipkit) , flush = True)  
            if os.path.isfile(out_dir + name2):
                os.remove(out_dir + name2)
            raise PhyloError

    else:
        with open(out_dir + '/clipkit.sh', 'w') as out:
            if '#PBS' in hpc:
                out.write( hpc + '\n\n' + #cmd1 + '\n' + 
                    ' '.join([str(x) for x in cmd]) + '\n\ncd ' + out_dir + '\nqsub tree.sh' )
            else:
                out.write( hpc + '\n\n' + #cmd1 + '\n' + 
                    ' '.join([str(x) for x in cmd]) + '\n\ncd ' + out_dir + '\nsbatch tree.sh' )
    
    return out_dir + '/' + name2


def run_mf(clipkit_file, out_dir, hpc, verbose, cpus = 1, spacer = '\t'):
    cmd = ['iqtree', '-m', 'MFP+MERGE', '-T', '1', '-B', '1000', '-s',
           clipkit_file]
    vprint(spacer + clipkit_file, v = verbose, flush = True)
    if verbose:
        run_mf = subprocess.call(cmd)
    else:
        run_mf = subprocess.call(cmd, stdout = subprocess.DEVNULL,
                                 stderr = subprocess.DEVNULL)
    with open(clipkit_file + '.log', 'r') as raw:
        for line in raw:
            if line.startswith('Best-fit model:'):
                model = re.search(r'Best-fit model: ([^\W]+)', line)[1]
                break

    return model

def prepare_nexus(concat_fa, models):

    nex_data = '#nexus\nbegin sets;'
    index = 1
    for i, trimmed_f in enumerate(list(models.keys())):
        trim_fa = fa2dict(trimmed_f)
        len0 = trim_fa[list(trim_fa.keys())[0]]['sequence']
        if not all(len(x['sequence']) == len0):
            eprint(spacer + 'ERROR: alignment sequences are not same length ' \
                   + trimmed_f, flush = True)
            sys.exit(17)
        new_index = index + len0
        nex_data += '\n    charset part' + str(i + 1) + ' = ' \
                  + str(index) + '-' + str(new_index) + ';'
        index = newindex

        for seq, data in trim_fa.items():
            ome = seq[:seq.find('_')]
            if ome in concat_fa:
                concatfa_data[ome]['sequence'] += data['sequence']

    nex_data += '\n    charpartition mine = '
    for i, model in models.values():
        nex_data += model + ':part' + str(i +1) + ','
    nex_data = nex_data[:-1] + ';\nend;'

    return nex_data, concat_fa 

def run_partition_tree(fa_file, nex_file, verbose, cpus, spacer = '\t'):
    cmd = ['iqtree', '-s', fa_file, '-p', nex_file, '-T', str(cpus),
           '-B', '1000', '--sampling', 'GENESITE']
    print(spacer + 'Tree building', flush = True)
    if verbose:
        run_tree = subprocess.call(cmd)
    else:
        run_tree = subprocess.call(cmd, stdout = subprocess.DEVNULL,
                                   stderr = subprocess.DEVNULL)
    return run_tree



def treeRun(clipkit_file, out_dir, hpc, verbose, fast = False, cpus = 1, spacer = '\t'):

    if fast:
        cmd = ['fasttree', '-out', clipkit_file + '.treefile', clipkit_file]
    else:
        cmd = ['iqtree', '-s', clipkit_file, '-B', '1000', '-T', str(cpus)]

   
    if not hpc:
        print(spacer + 'Tree building', flush = True)
        if verbose:
            run_tree = subprocess.call(
                cmd
                )
        else:
            run_tree = subprocess.call(
                cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
                )
        if run_tree != 0:
            eprint(spacer + '\tERROR: tree failed: ' + str(run_tree), flush = True)
            raise PhyloError
    else: 
        vprint('\nOutputting bash script `tree.sh`.\n', v = verbose, flush = True)

        with open(out_dir + '/tree.sh', 'w') as out:
            out.write(hpc + '\n\n' + ' '.join([str(x) for x in cmd]))


def main( 
    fasta_path, slurm = False, torque = False, fast = False, project = '', 
    output_dir = None, verbose = True, alignment = False, mem = '60GB', cpus = 1,
    spacer = '\t\t', align_stop = False, tree_stop = False, 
    flag_incomplete = True
    ):

    hpc = False
    if slurm:
        vprint('\nHPC mode, preparing submission scripts (Slurm)\n', v = verbose, flush = True)
        hpc = '#!/bin/bash\n#SBATCH --time=24:00:00\n#SBATCH --nodes=1\n' + \
            '#SBATCH --ntasks-per-node=' + str(cpus) + '\n#SBATCH -A ' + project + '\n' + \
            '#SBATCH --mem=' + str(mem)
#            '\n\nsource activate ' + source
    elif torque:
        vprint('\nHPC mode, preparing submission scripts (PBS)\n', v = verbose, flush = True)
        hpc = '#PBS -l walltime=10:00:00\n#PBS -l nodes=1:ppn=4\n#PBS -A ' + project

    partition = False
    output_dir_prep = os.getcwd() + '/'
    if len(fasta_path) == 1:
        if os.path.isfile(fasta_path):
            if not output_dir:
                dir_name = re.sub(r'\..*?$', '_tree', os.path.basename( os.path.abspath(fasta_path)))
                out_dir = output_dir + '/' + dir_name
                if not os.path.isdir(out_dir):
                    os.mkdir(out_dir)
                out_dir = format_path(out_dir)
            else:
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)
                out_dir = format_path(output_dir)
            fastas = [fasta_path]
        elif os.path.isdir(fasta_path):
            vprint('\nDirectory input, partition analysis mode', 
                   v = verbose, flush = True)
            partition = True
            if not output_dir:
                out_dir = mkOutput(output_dir_prep, 'fa2tree')
            else:
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)
                out_dir = format_path(output_dir)
            fastas = collect_files(fasta_path, 'fa')
            fastas.extend(collect_files(fasta_path, 'faa'))
            fastas.extend(collect_files(fasta_path, 'fna'))
            fastas.extend(collect_files(fasta_path, 'fasta'))
            fastas.extend(collect_files(fasta_path, 'fsa'))
            if prealigned:
                fastas.extend(collect_files(fasta_path, 'mafft'))
                fastas.extend(collect_files(fasta_path, 'clipkit'))
                check_fas = set(fastas)
                new_fas = []
                for fa in fastas:
                    if fa + '.clipkit' not in check_fas:
                        if fa + '.mafft' not in check_fas:
                            new_fas.append(fa)
                fastas = new_fas
    elif len(fasta_path) > 1:
        vprint('\nMultiple fastas inputted, partition analysis mode',
               v = verbose, flush = True)
        partition = True
        if not output_dir:
            out_dir = mkOutput(output_dir_prep, 'fa2tree')
        else:
            if not os.path.isdir(output_dir):
                os.mkdir(output_dir)
            out_dir = format_path(output_dir)
        fastas = [format_path(x) for x in fasta_path]

    if partition: #multigene partition mode
        ome2fa2gene = defaultdict(dict)
        complete_omes = []
        for fasta in fastas:
            fa_dict = fa2dict(fasta)
            for seq in fa_dict:
                ome = seq[:seq.find('_')]
                if fasta not in ome2fa2gene[ome]:
                    ome2fa2gene[ome][fasta] = seq
                else:
                    eprint('\nERROR: multiple sequences for a \
                            single ome in ' + fasta, flush = True)
                    sys.exit(5)
            incomp_omes = set([x for x, fas in ome2fa2gene.items() \
                                 if len(fas) != len(fastas)])
            if incomp_omes:
                if flag_incomplete:
                    eprint('\nERROR: omes without sequences in all fastas: ' \
                         + ','.join([str(x) for x in list(incomp_omes)]))
                    sys.exit(6)
                else:
                    eprint('\nWARNING: omes removed without sequences in all \
                            fastas: ' \
                         + ','.join([str(x) for x in list(incomp_omes)]))

    trimmed_files = []
    for fasta in fastas:
        if not prealigned:
            fasta_name = os.path.basename(os.path.abspath(fasta))
            mafft = out_dir + '/' + fasta_name + '.mafft'
            clipkit = out_dir + '/' + fasta_name + '.clipkit'
            
            try:
                if os.stat(mafft):
                    vprint('\nAlignment exists', v = verbose, flush = True)
                else:
                    raise ValueError
            except (FileNotFoundError, ValueError) as e:
                if not alignment:
                    mafft = mafftRun(os.path.abspath(fasta_path), out_dir, hpc, 
                                     verbose, cpus, spacer = spacer)
                else:
                    mafft = fasta_path
        
            try:
                if os.stat(clipkit):
                    vprint('\nTrim exists', v = verbose, flush = True)
                    clipkit_out = clipkit
                else:
                    raise ValueError
            except (FileNotFoundError, ValueError) as e:
                clipkit_out = trimRun(mafft, out_dir, hpc, verbose, spacer = spacer)
            trimmed_files.append(clipkit_out)

    if align_stop:
        vprint(spacer + 'Alignments outputed', v = verbose, flush = True)
        sys.exit(0)

    if not partition:
        for trimmed_f in trimmed_files:
            treeRun(trimmed_f, out_dir, hpc, verbose, fast, cpus, spacer = spacer)
    else:
        print(spacer + 'Model finding', flush = True)
        models = {}
        for trimmed_f in trimmed_files:
            model = run_mf(trimmed_f, out_dir, hpc, verbose, cpus,
                   spacer = spacer + '\t')
            models[trimmed_f] = model

        print(spacer + 'Concatenating', flush = True)
        concat_fa = {ome: {'description': '', 'sequence': ''}
                     for ome in ome2fa2gene if ome not in incomp_omes}
        nex_data, concat_fa = prepare_nexus(concat_fa, models)
        with open(out_dir + 'concatenated.fa', 'w') as out:
            out.write(dict2fa(concat_fa))
        with open(out_dir + 'concatenated.nex', 'w') as out:
            out.write(nex_data)

        if tree_stop:
            vprint(spacer + 'Concatenated nexus and fasta outputted', 
                    v = verbose, flush = True)
            sys.exit(0)

        run_partition_tree(out_dir + 'concatenated.fa', 
                           out_dir + 'concatenated.nex',
                           verbose, cpus, spacer = spacer)

    if hpc:
        if slurm:
            vprint('\nStart pipeline via `sbatch <STARTSTEP>.sh` in ' + out_dir + '\n',
                v = verbose)
        else:
            vprint('\nStart pipeline via `qsub <STARTSTEP>.sh` in ' + out_dir + '\n',
                v = verbose)


if __name__ == '__main__':

    parser = argparse.ArgumentParser( 
        description = 'Takes in a multifasta, or directory/list of fastas [partition]  ' + \
        ' aligns (mafft), trims (clipkit), and infers phylogeny (iqtree/fastree).'
        )

    io_opt = parser.add_argument_group('Input/Output')
    io_opt.add_argument('-i', '--input', required = True,
        help = 'Fasta or directory of fastas/alignments [partition]. \
                "-" for stdin')
    io_opt.add_argument('-o', '--output')

    s_opt = parser.add_argument_group('Single tree options')
    s_opt.add_argument('--fast', action = 'store_true', \
        help = 'Fasttree' )
    s_opt.add_argument('-s', '--slurm', action = 'store_true', \
        help = 'Submit ALL steps to Slurm' )
    s_opt.add_argument('-t', '--torque', action = 'store_true', \
        help = 'Submit ALL steps to Torque.' )
    s_opt.add_argument('-A', '--project', help = 'HPC project')

    m_opt = parser.add_argument_group('Partition tree options')
    m_opt.add_argument('-a', '--stop_align', action = 'store_true',
        help = 'Stop after aligning and trimming')
    m_opt.add_argument('-p', '--prealigned', action = 'store_true',
        help = 'Use alignments')
    m_opt.add_argument('-c', '--stop_cat', action = 'store_true',
        help = 'Stop after concatenating sequences')
    m_opt.add_argument('-m', '--missing', action = 'store_true',
        help = 'Remove omes with missing sequences')
    args = parser.parse_args()

    output = format_path(args.output)
    if output:
        if not output.endswith('/'):
            output += '/' # bring to mycotools path expectations
    args_dict = {
        'Fasta': args.input, 'Fast': args.fast, 'Prealigned': args.prealigned,
        'Alignment stop': args.stop_align, 'Concatenation stop': args.stop_cat,
        'Ignore missing': args.missing,
        'Torque': args.torque, 'Slurm': args.slurm, 'HPC project': args.project, 'Output': output
        }
    start_time = intro('Fasta2Tree', args_dict)

    execs = ['mafft', 'clipkit']
    if args.fast:
        execs.append('fasttree')
    else:
        execs.append('iqtree')
    if args.torque or args.slurm:
        findExecs(execs, execs)
    else:
        findExecs(execs, execs)

    if args.input == '-':
        args.input = \
            stdin2str().replace('"','').replace("'",'').split()
    else:
        args.input = args.input.replace('"','').replace("'",'').split()
    main( 
        args.input, slurm = args.slurm, fast = args.fast,
        torque = args.torque, project = args.project,
        output_dir = output, verbose = True, alignment = args.prealigned,
        align_stop = args.stop_align, tree_stop = args.stop_cat,
        flag_incomplete = bool(not args.missing)
        )
#    elif os.path.isdir(input_check):
 #       fas = collect_files(input_check, 'fa')
  #      fas.extend(collect_files( input_check, 'fasta' ))
#        fas.extend(collect_files( input_check, 'fna' ))
 #       fas.extend(collect_files( input_check, 'fsa' ))
  #      if not fas:
   #         eprint('\nERROR: no {.fa, .fasta, .fna, .fsa} detected', flush = True)
    #        sys.exit(2)
     #   print(flush = True)
      #  for fa in fas:
       #     print(fa, flush = True)
        #    main(
         #       fa, slurm = args.slurm, torque = args.torque, alignment = alignment, fast = args.fast,
          #      project = args.project, output_dir = output, verbose = False, spacer = '\n'
           #     )
#    else:
 #       eprint('\nERROR: invalid input', flush = True)
  #      sys.exit(3)

    outro(start_time)
